from collections.abc import Callable
import time
from typing import Any, Literal

from azure.identity import DefaultAzureCredential, get_bearer_token_provider
from openai import AzureOpenAI, OpenAI

from not_again_ai.llm.embedding.types import EmbeddingObject, EmbeddingRequest, EmbeddingResponse


def openai_create_embeddings(request: EmbeddingRequest, client: Callable[..., Any]) -> EmbeddingResponse:
    kwargs = request.model_dump(mode="json", exclude_none=True)

    start_time = time.time()
    response = client(**kwargs)
    end_time = time.time()
    response_duration = round(end_time - start_time, 4)

    embeddings: list[EmbeddingObject] = []
    for data in response["data"]:
        embeddings.append(EmbeddingObject(embedding=data["embedding"], index=data["index"]))

    return EmbeddingResponse(
        embeddings=embeddings,
        response_duration=response_duration,
        total_tokens=response["usage"]["total_tokens"],
    )


def create_client_callable(client_class: type[OpenAI | AzureOpenAI], **client_args: Any) -> Callable[..., Any]:
    """Creates a callable that instantiates and uses an OpenAI client.

    Args:
        client_class: The OpenAI client class to instantiate (OpenAI or AzureOpenAI)
        **client_args: Arguments to pass to the client constructor

    Returns:
        A callable that creates a client and returns completion results
    """
    filtered_args = {k: v for k, v in client_args.items() if v is not None}

    def client_callable(**kwargs: Any) -> Any:
        client = client_class(**filtered_args)
        completion = client.embeddings.create(**kwargs)
        return completion.to_dict()

    return client_callable


class InvalidOAIAPITypeError(Exception):
    """Raised when an invalid OAIAPIType string is provided."""


def openai_client(
    api_type: Literal["openai", "azure_openai"] = "openai",
    api_key: str | None = None,
    organization: str | None = None,
    aoai_api_version: str = "2024-06-01",
    azure_endpoint: str | None = None,
    timeout: float | None = None,
    max_retries: int | None = None,
) -> Callable[..., Any]:
    """Create an OpenAI or Azure OpenAI client instance based on the specified API type and other provided parameters.

    It is preferred to use RBAC authentication for Azure OpenAI. You must be signed in with the Azure CLI and have correct role assigned.
    See https://techcommunity.microsoft.com/t5/microsoft-developer-community/using-keyless-authentication-with-azure-openai/ba-p/4111521

    Args:
        api_type (str, optional): Type of the API to be used. Accepted values are 'openai' or 'azure_openai'.
            Defaults to 'openai'.
        api_key (str, optional): The API key to authenticate the client. If not provided,
            OpenAI automatically uses `OPENAI_API_KEY` from the environment.
            If provided for Azure OpenAI, it will be used for authentication instead of the Azure AD token provider.
        organization (str, optional): The ID of the organization. If not provided,
            OpenAI automotically uses `OPENAI_ORG_ID` from the environment.
        aoai_api_version (str, optional): Only applicable if using Azure OpenAI https://learn.microsoft.com/azure/ai-services/openai/reference#rest-api-versioning
        azure_endpoint (str, optional): The endpoint to use for Azure OpenAI.
        timeout (float, optional): By default requests time out after 10 minutes.
        max_retries (int, optional): Certain errors are automatically retried 2 times by default,
            with a short exponential backoff. Connection errors (for example, due to a network connectivity problem),
            408 Request Timeout, 409 Conflict, 429 Rate Limit, and >=500 Internal errors are all retried by default.

    Returns:
        Callable[..., Any]: A callable that creates a client and returns completion results


    Raises:
        InvalidOAIAPITypeError: If an invalid API type string is provided.
        NotImplementedError: If the specified API type is recognized but not yet supported (e.g., 'azure_openai').
    """
    if api_type not in ["openai", "azure_openai"]:
        raise InvalidOAIAPITypeError(f"Invalid OAIAPIType: {api_type}. Must be 'openai' or 'azure_openai'.")

    if api_type == "openai":
        return create_client_callable(
            OpenAI,
            api_key=api_key,
            organization=organization,
            timeout=timeout,
            max_retries=max_retries,
        )
    elif api_type == "azure_openai":
        if api_key:
            return create_client_callable(
                AzureOpenAI,
                api_version=aoai_api_version,
                azure_endpoint=azure_endpoint,
                api_key=api_key,
                timeout=timeout,
                max_retries=max_retries,
            )
        else:
            azure_credential = DefaultAzureCredential()
            ad_token_provider = get_bearer_token_provider(
                azure_credential, "https://cognitiveservices.azure.com/.default"
            )
            return create_client_callable(
                AzureOpenAI,
                api_version=aoai_api_version,
                azure_endpoint=azure_endpoint,
                azure_ad_token_provider=ad_token_provider,
                timeout=timeout,
                max_retries=max_retries,
            )
    else:
        raise NotImplementedError(f"API type '{api_type}' is invalid.")



from abc import ABC, abstractmethod
from collections.abc import Collection, Set
from typing import Literal

from not_again_ai.llm.chat_completion.types import MessageT


class BaseTokenizer(ABC):
    def __init__(
        self,
        model: str,
        provider: str,
        allowed_special: Literal["all"] | Set[str] | None = None,
        disallowed_special: Literal["all"] | Collection[str] | None = None,
    ):
        self.model = model
        self.provider = provider
        self.allowed_special = allowed_special
        self.disallowed_special = disallowed_special

        self.init_tokenizer(model, provider, allowed_special, disallowed_special)

    @abstractmethod
    def init_tokenizer(
        self,
        model: str,
        provider: str,
        allowed_special: Literal["all"] | Set[str] | None = None,
        disallowed_special: Literal["all"] | Collection[str] | None = None,
    ) -> None:
        pass

    @abstractmethod
    def truncate_str(self, text: str, max_len: int) -> str:
        pass

    @abstractmethod
    def num_tokens_in_str(self, text: str) -> int:
        pass

    @abstractmethod
    def num_tokens_in_messages(self, messages: list[MessageT]) -> int:
        pass



from collections.abc import Callable
from typing import Any

import pytest

from not_again_ai.llm.embedding import EmbeddingRequest, create_embeddings
from not_again_ai.llm.embedding.providers.ollama_api import ollama_client
from not_again_ai.llm.embedding.providers.openai_api import openai_client
from not_again_ai.llm.embedding.types import EmbeddingResponse


def print_embedding_response(embedding_response: EmbeddingResponse, max_elements: int = 5) -> None:
    """Print an EmbeddingResponse, truncating each embedding vector after `max_elements` values.

    Args:
        embedding_response (EmbeddingResponse): The response to print.
        max_elements (int): Maximum number of elements from each embedding vector to display.
                            Defaults to 5.
    """
    print("EmbeddingResponse:")
    print(f"  Response Duration: {embedding_response.response_duration}")
    print(f"  Total Tokens: {embedding_response.total_tokens}")
    print(f"  Extras: {embedding_response.extras}")
    print("  Embeddings:")
    for obj in embedding_response.embeddings:
        vector = obj.embedding
        if len(vector) > max_elements:
            # Format the first max_elements, then append an ellipsis.
            vector_str = ", ".join(f"{value:.4f}" for value in vector[:max_elements])
            vector_str += ", ..."
        else:
            vector_str = ", ".join(f"{value:.4f}" for value in vector)
        print(f"    Index {obj.index}: [{vector_str}]")


# region OpenAI and Azure OpenAI Embedding
@pytest.fixture(
    params=[
        {},
        {"api_type": "azure_openai", "aoai_api_version": "2025-01-01-preview"},
    ]
)
def openai_aoai_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return openai_client(**request.param)


def test_create_embeddings(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = EmbeddingRequest(input="Hello, world!", model="text-embedding-3-small")
    response = create_embeddings(request, "openai", openai_aoai_client_fixture)
    print_embedding_response(response)


def test_create_embeddings_multiple(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = EmbeddingRequest(input=["Hello, world!", "Hello, world 2!"], model="text-embedding-3-small")
    response = create_embeddings(request, "openai", openai_aoai_client_fixture)
    print_embedding_response(response)


def test_create_embeddings_dimensions(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = EmbeddingRequest(
        input="This is a test of the dimensions parameter",
        model="text-embedding-3-large",
        dimensions=3,
    )
    response = create_embeddings(request, "openai", openai_aoai_client_fixture)
    print_embedding_response(response)


# endregion
# region Ollama
@pytest.fixture(
    params=[
        {},
    ]
)
def ollama_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return ollama_client(**request.param)


def test_create_embeddings_ollama(ollama_client_fixture: Callable[..., Any]) -> None:
    request = EmbeddingRequest(input="Hello, world!", model="snowflake-arctic-embed2")
    response = create_embeddings(request, "ollama", ollama_client_fixture)
    print_embedding_response(response)


def test_create_embeddings_ollama_missing_param(ollama_client_fixture: Callable[..., Any]) -> None:
    request = EmbeddingRequest(input="Hello, world!", model="snowflake-arctic-embed2", dimensions=3)
    response = create_embeddings(request, "ollama", ollama_client_fixture)
    print_embedding_response(response)


def test_create_embeddings_ollama_multiple(ollama_client_fixture: Callable[..., Any]) -> None:
    request = EmbeddingRequest(input=["Hello, world!", "Hello, world 2!"], model="snowflake-arctic-embed2")
    response = create_embeddings(request, "ollama", ollama_client_fixture)
    print_embedding_response(response)


# endregion



import numpy as np
import numpy.typing as npt
import scipy
import sklearn.metrics as skmetrics
import sklearn.model_selection as skmodel_selection
import sklearn.preprocessing as skpreprocessing
import sklearn.tree as sktree


def _process_variable(
    x: npt.NDArray[np.int_] | (npt.NDArray[np.float64] | npt.NDArray[np.str_]),
) -> npt.NDArray[np.int_] | (npt.NDArray[np.float64] | npt.NDArray[np.str_]):
    """Process variable by encoding it as a numeric array."""
    le = skpreprocessing.LabelEncoder()
    x = le.fit_transform(x)
    return x


def pearson_correlation(
    x: list[int]
    | (list[float] | (list[str] | (npt.NDArray[np.int_] | (npt.NDArray[np.float64] | npt.NDArray[np.str_])))),
    y: list[int]
    | (list[float] | (list[str] | (npt.NDArray[np.int_] | (npt.NDArray[np.float64] | npt.NDArray[np.str_])))),
    is_x_categorical: bool = False,
    is_y_categorical: bool = False,
    print_diagnostics: bool = False,
) -> float:
    """Absolute value of the Pearson correlation coefficient.
    Returns 1 in the case y contains all of the same values.

    Implemented using scipy https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.pearsonr.html

    Args:
        x (listlike): first variable
        y (listlike): second variable
        is_x_categorical (bool): whether x is categorical
        is_y_categorical (bool): whether y is categorical
        print_diagnostics (bool): whether to print diagnostics to stdout
    """
    x_array = np.array(x)
    y_array = np.array(y)

    if is_x_categorical:
        x_array = _process_variable(x_array)

    if is_y_categorical:
        y_array = _process_variable(y_array)

    # check if y contains all of the same values, and if so return 1
    if len(np.unique(y_array)) == 1:
        if print_diagnostics:
            print("y contains all of the same values, returning 1")
        return 1.0

    pearsonr = scipy.stats.pearsonr(x_array, y_array)
    metric: float = pearsonr.statistic
    metric = np.abs(metric)
    return metric


def pred_power_score_classification(
    x: list[int]
    | (list[float] | (list[str] | (npt.NDArray[np.int_] | (npt.NDArray[np.float64] | npt.NDArray[np.str_])))),
    y: list[int] | (list[str] | npt.NDArray[np.int_]),
    cv_splits: int = 5,
    print_diagnostics: bool = False,
) -> float:
    """Compute Predictive Power Score, an asymmetric score that can detect
    linear or non-linear relationships between two variables.
    For this implementation, the score is computed for a classification task and y must be categorical.

    Returns 1 in the case y contains all of the same values.

    Args:
        x (listlike of int, float, or string): first variable
        y (listlike of int or string): second variable
        cv_splits (int): number of cross-validation splits
        print_diagnostics (bool): whether to print diagnostics to stdout
    """
    x_array = np.array(x)
    y_array = np.array(y)

    le = skpreprocessing.LabelEncoder()
    # check if x contains any strings
    if any(isinstance(elem, str) for elem in x_array):
        x_array = le.fit_transform(x_array)

    x_array = x_array.reshape(-1, 1)
    y_array = le.fit_transform(y_array)

    # check if y contains all of the same values, and if so return 1
    if len(np.unique(y_array)) == 1:
        if print_diagnostics:
            print("y contains all of the same values, returning 1")
        return 1.0

    # Use KFold cross-validation to compute weighted (macro) F1 score
    model = sktree.DecisionTreeClassifier(criterion="gini", splitter="best", max_depth=None, random_state=0)
    cv_method = skmodel_selection.KFold(n_splits=cv_splits, shuffle=True, random_state=0)
    f1_scores = skmodel_selection.cross_val_score(
        model, x_array, y_array, cv=cv_method, scoring="f1_weighted", error_score="raise"
    )
    f1 = np.mean(f1_scores)

    # find majority class in y
    majority_class = np.argmax(np.bincount(y_array))
    preds = np.ones_like(y_array) * majority_class
    f1_null: float = skmetrics.f1_score(y_array, preds, average="weighted")

    # generate random predictions
    preds = np.random.choice(np.unique(y_array), size=len(y_array))
    f1_random: float = skmetrics.f1_score(y_array, preds, average="weighted")

    f1_naive = np.max([f1_null, f1_random])
    pps: float = (f1 - f1_naive) / (1 - f1_naive)

    # ensure pps is not negative
    pps = np.max([0, pps])
    return pps



from collections.abc import Callable
import os
import re
import time
from typing import Any

from loguru import logger
from ollama import Client, EmbedResponse, ResponseError

from not_again_ai.llm.embedding.types import EmbeddingObject, EmbeddingRequest, EmbeddingResponse

OLLAMA_PARAMETER_MAP = {
    "dimensions": None,
}


def validate(request: EmbeddingRequest) -> None:
    # Check if any of the parameters set to OLLAMA_PARAMETER_MAP are not None
    for key, value in OLLAMA_PARAMETER_MAP.items():
        if value is None and getattr(request, key) is not None:
            logger.warning(f"Parameter {key} is not supported by Ollama and will be ignored.")


def ollama_create_embeddings(request: EmbeddingRequest, client: Callable[..., Any]) -> EmbeddingResponse:
    validate(request)
    kwargs = request.model_dump(mode="json", exclude_none=True)

    # For each key in OLLAMA_PARAMETER_MAP
    # If it is not None, set the key in kwargs to the value of the corresponding value in OLLAMA_PARAMETER_MAP
    # If it is None, remove that key from kwargs
    for key, value in OLLAMA_PARAMETER_MAP.items():
        if value is not None and key in kwargs:
            kwargs[value] = kwargs.pop(key)
        elif value is None and key in kwargs:
            del kwargs[key]

    # Explicitly set truncate to True (it is the default)
    kwargs["truncate"] = True

    try:
        start_time = time.time()
        response: EmbedResponse = client(**kwargs)
        end_time = time.time()
        response_duration = round(end_time - start_time, 4)
    except ResponseError as e:
        # If the error says "model 'model' not found" use regex then raise a more specific error
        expected_pattern = f"model '{request.model}' not found"
        if re.search(expected_pattern, e.error):
            raise ResponseError(f"Model '{request.model}' not found.") from e
        else:
            raise ResponseError(e.error) from e

    embeddings: list[EmbeddingObject] = []
    for index, embedding in enumerate(response.embeddings):
        embeddings.append(EmbeddingObject(embedding=list(embedding), index=index))

    return EmbeddingResponse(
        embeddings=embeddings,
        response_duration=response_duration,
        total_tokens=response.prompt_eval_count,
    )


def ollama_client(host: str | None = None, timeout: float | None = None) -> Callable[..., Any]:
    """Create an Ollama client instance based on the specified host or will read from the OLLAMA_HOST environment variable.

    Args:
        host (str, optional): The host URL of the Ollama server.
        timeout (float, optional): The timeout for requests

    Returns:
        Client: An instance of the Ollama client.

    Examples:
        >>> client = client(host="http://localhost:11434")
    """
    if host is None:
        host = os.getenv("OLLAMA_HOST")
        if host is None:
            logger.warning("OLLAMA_HOST environment variable not set, using default host: http://localhost:11434")
            host = "http://localhost:11434"

    def client_callable(**kwargs: Any) -> Any:
        client = Client(host=host, timeout=timeout)
        return client.embed(**kwargs)

    return client_callable



from collections.abc import Collection, Set
from typing import Literal

from loguru import logger

from not_again_ai.llm.chat_completion.types import MessageT
from not_again_ai.llm.prompting.providers.openai_tiktoken import TokenizerOpenAI
from not_again_ai.llm.prompting.types import BaseTokenizer


class Tokenizer(BaseTokenizer):
    def __init__(
        self,
        model: str,
        provider: str,
        allowed_special: Literal["all"] | Set[str] | None = None,
        disallowed_special: Literal["all"] | Collection[str] | None = None,
    ):
        self.model = model
        self.provider = provider
        self.allowed_special = allowed_special
        self.disallowed_special = disallowed_special

        self.init_tokenizer(model, provider, allowed_special, disallowed_special)

    def init_tokenizer(
        self,
        model: str,
        provider: str,
        allowed_special: Literal["all"] | Set[str] | None = None,
        disallowed_special: Literal["all"] | Collection[str] | None = None,
    ) -> None:
        if provider == "openai" or provider == "azure_openai":
            self.tokenizer = TokenizerOpenAI(model, provider, allowed_special, disallowed_special)
        else:
            logger.warning(f"Provider {provider} not supported. Initializing using tiktoken and gpt-4o.")
            self.tokenizer = TokenizerOpenAI("gpt-4o", "openai", allowed_special, disallowed_special)

    def truncate_str(self, text: str, max_len: int) -> str:
        return self.tokenizer.truncate_str(text, max_len)

    def num_tokens_in_str(self, text: str) -> int:
        return self.tokenizer.num_tokens_in_str(text)

    def num_tokens_in_messages(self, messages: list[MessageT]) -> int:
        return self.tokenizer.num_tokens_in_messages(messages)



from collections.abc import Callable
import os
from pathlib import Path
from typing import Any

import pytest

from not_again_ai.llm.chat_completion import chat_completion
from not_again_ai.llm.chat_completion.providers.anthropic_api import anthropic_client
from not_again_ai.llm.chat_completion.providers.ollama_api import ollama_client
from not_again_ai.llm.chat_completion.providers.openai_api import openai_client
from not_again_ai.llm.chat_completion.types import (
    AssistantMessage,
    ChatCompletionRequest,
    Function,
    ImageContent,
    ImageDetail,
    ImageUrl,
    MessageT,
    SystemMessage,
    TextContent,
    ToolCall,
    ToolMessage,
    UserMessage,
)
from not_again_ai.llm.prompting.compile_prompt import encode_image

image_dir = Path(__file__).parent.parent / "sample_images"
cat_image = image_dir / "cat.jpg"
dog_image = image_dir / "dog.jpg"
numbers_image = image_dir / "numbers.png"
sk_infographic = image_dir / "SKInfographic.png"
sk_diagram = image_dir / "SKDiagram.png"


# region OpenAI and Azure OpenAI Chat Completion
@pytest.fixture(
    params=[
        {},
        {"api_type": "azure_openai", "aoai_api_version": "2025-01-01-preview"},
    ]
)
def openai_aoai_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return openai_client(**request.param)


def test_chat_completion_simple(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="Hello, world!"),
            UserMessage(content="What is the capital of France?"),
        ],
        max_completion_tokens=100,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_length(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-2024-11-20",
        messages=[
            SystemMessage(content="Hello, world!"),
            UserMessage(content="What is the capital of France and the capital of Germany?"),
        ],
        max_completion_tokens=2,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_json_mode(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(
            content="""You are getting names of users and formatting them into json.
Example:
User: Jane Doe
Output: {"name": "Jane Doe"}"""
        ),
        UserMessage(content="John Doe"),
    ]
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=messages,
        max_completion_tokens=300,
        temperature=0,
        json_mode=True,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_n(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="You are a helpful assistant."),
            UserMessage(content="Hello!"),
        ],
        max_completion_tokens=100,
        n=2,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_seed(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="You are a helpful assistant."),
            UserMessage(content="Generate a random number between 0 and 100."),
        ],
        max_completion_tokens=100,
        temperature=2,
        seed=42,
    )

    response_1 = chat_completion(request, "openai", openai_aoai_client_fixture)
    response_2 = chat_completion(request, "openai", openai_aoai_client_fixture)

    print(response_1.choices[0].message.content)
    print(response_2.choices[0].message.content)


def test_chat_completion_logprobs(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="You are a helpful assistant."),
            UserMessage(content="Hello!"),
        ],
        max_completion_tokens=100,
        logprobs=True,
        top_logprobs=None,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_toplogprobs(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="You are a helpful assistant."),
            UserMessage(content="Hello!"),
        ],
        max_completion_tokens=100,
        logprobs=True,
        top_logprobs=3,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_tool_simple(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        }
    ]

    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            UserMessage(
                content="What's the current weather like in Boston, MA today? Call the get_current_weather function."
            )
        ],
        tools=tools,
        max_completion_tokens=300,
        temperature=0,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_tool_required_name(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            UserMessage(content="What's the current weather like in Boston, MA today?"),
        ],
        tools=tools,
        tool_choice="get_current_weather",
        max_completion_tokens=300,
        temperature=0,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_required_tool_call(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(
            content="You are a helpful assistant who always listens to the user and does not call any tools."
        ),
        UserMessage(content="Do not call any tools!"),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        tools=tools,
        tool_choice="required",
        max_completion_tokens=400,
        temperature=0.5,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_multiple_tools(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(content="Call the get_current_weather function once for each city that the user mentions."),
        UserMessage(content="What's the current weather like in Boston, MA and New York, NY today?"),
    ]
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=messages,
        tools=tools,
        max_completion_tokens=400,
        temperature=0,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_tool_choice_none(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(content="Call the get_current_weather function once for each city that the user mentions."),
        UserMessage(content="What's the current weather like in Boston, MA and New York, NY today?"),
    ]
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=messages,
        tools=tools,
        tool_choice="none",
        max_completion_tokens=400,
        temperature=0,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_message_with_tools(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(
            content="""You will be given a function called get_current_weather.
Before calling the function, first reason about which city the user is asking about. YOU MUST think step by step before calling the function.
For example, if the user asks 'What's the current weather like in Boston, MA today?', You should first say 'The user is asking about Boston, MA so I will call the function with 'Boston, MA' <now call the function>"""
        ),
        UserMessage(
            content="What's the current weather like in Boston, MA today? First think step by step as to which city and state to call, only then call the get_current_weather function. "
        ),
    ]

    request = ChatCompletionRequest(
        model="gpt-4o-2024-11-20",
        messages=messages,
        tools=tools,
        max_completion_tokens=600,
        temperature=0.7,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_tool_message(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant"),
        UserMessage(content="What is the weather in Boston, MA?"),
        AssistantMessage(
            content="",
            tool_calls=[
                ToolCall(
                    id="abc123",
                    function=Function(
                        name="get_current_weather",
                        arguments={"location": "Boston, MA"},
                    ),
                )
            ],
        ),
        ToolMessage(name="abc123", content="The weather in Boston, MA is 70 degrees Fahrenheit."),
    ]
    request = ChatCompletionRequest(
        model="gpt-4o-2024-11-20",
        messages=messages,
        max_completion_tokens=300,
        temperature=0.3,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_structured_output(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant"),
        UserMessage(content="9.11 and 9.9 -- which is bigger?"),
    ]
    json_schema = {
        "name": "reasoning_schema",
        "strict": True,
        "schema": {
            "type": "object",
            "properties": {
                "reasoning_steps": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The reasoning steps leading to the final conclusion.",
                },
                "answer": {
                    "type": "string",
                    "description": "The final answer, taking into account the reasoning steps.",
                },
            },
            "required": ["reasoning_steps", "answer"],
            "additionalProperties": False,
        },
        "description": "A schema for structured output that includes reasoning steps and the final answer.",
    }

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=400,
        temperature=0.2,
        json_mode=False,
        structured_outputs=json_schema,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_required_tools_none_called_structured(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "strict": True,
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                    "additionalProperties": False,
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(content="Do not call get_current_weather or any other tool under any circumstances."),
        UserMessage(content="What is 2+2?"),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        tools=tools,
        tool_choice="required",
        max_completion_tokens=300,
        temperature=0.5,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_vision(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=200,
        temperature=0.5,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_vision_n(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]
    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=200,
        n=2,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_vision_seed(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="Pick one random number that is written in the image. Just write the number."),
                ImageContent(
                    image_url=ImageUrl(
                        url=f"data:image/png;base64,{encode_image(numbers_image)}", detail=ImageDetail.LOW
                    )
                ),
            ]
        ),
    ]
    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=100,
        temperature=2,
        seed=42,
    )

    response_1 = chat_completion(request, "openai", openai_aoai_client_fixture)
    response_2 = chat_completion(request, "openai", openai_aoai_client_fixture)

    print(response_1.model_dump(mode="json", exclude_none=True))
    print(response_1.choices[0].message.content)
    print(response_2.choices[0].message.content)


def test_chat_completion_vision_multiple_images(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="What are the animals in the images? Reply in one word for each animal."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(dog_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]
    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=100,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_vision_multiple_messages(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    """Test with two image messages separated by an assistant message."""
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
        AssistantMessage(content="Cat"),
        UserMessage(
            content=[
                TextContent(text="What about this animal?"),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(dog_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=100,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_vision_different_fidelity(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    """Test sending one image with low fidelity and another with high fidelity."""
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(
                    text="Based on these infographics, can you summarize how Semantic Kernel works in exactly one sentence?"
                ),
                ImageContent(
                    image_url=ImageUrl(
                        url=f"data:image/png;base64,{encode_image(sk_infographic)}", detail=ImageDetail.HIGH
                    )
                ),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/png;base64,{encode_image(sk_diagram)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=200,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_vision_tool_call(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(
            content="""You are detecting if there is text (numbers or letters) in images. 
If you see any text, call the ocr tool. It takes no parameters."""
        ),
        UserMessage(
            content=[
                ImageContent(
                    image_url=ImageUrl(
                        url=f"data:image/png;base64,{encode_image(numbers_image)}", detail=ImageDetail.LOW
                    )
                ),
            ]
        ),
    ]
    tools = [
        {
            "type": "function",
            "function": {
                "name": "ocr",
                "description": "Perform Optical Character Recognition (OCR) on an image",
                "parameters": {},
            },
        },
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        tools=tools,
        max_completion_tokens=200,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_vision_json_mode(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(
            content="""You are performing optical character recognition (OCR) on an image. 
Return the recognized text in JSON format where the detected text is the value of the 'text' key."""
        ),
        UserMessage(
            content=[
                ImageContent(
                    image_url=ImageUrl(
                        url=f"data:image/png;base64,{encode_image(numbers_image)}", detail=ImageDetail.LOW
                    )
                ),
            ]
        ),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=300,
        json_mode=True,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_vision_many_features(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=100,
        temperature=2,
        n=2,
        logprobs=True,
        top_logprobs=2,
        seed=21,
    )

    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_invalid_params(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-2024-11-20",
        messages=[UserMessage(content="What is the capital of France?")],
        max_completion_tokens=100,
        context_window=1000,
        mirostat=1,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_max_tokens(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[UserMessage(content="What is the capital of France?")],
        max_tokens=100,
    )
    response = chat_completion(request, "openai", openai_aoai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


# region OpenAI
@pytest.fixture(
    params=[
        {},
    ]
)
def openai_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return openai_client(**request.param)


def test_chat_completion_o1_mini(openai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        UserMessage(
            content="You only think for one sentence, and then write at most a one sentence response. What is the capital of France?"
        ),
    ]
    request = ChatCompletionRequest(
        model="o1-mini-2024-09-12",
        messages=messages,
        max_completion_tokens=500,
    )
    response = chat_completion(request, "openai", openai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


# endregion


# region Azure OpenAI
@pytest.fixture(
    params=[
        {"api_type": "azure_openai", "aoai_api_version": "2024-10-01-preview"},
    ]
)
def azure_openai_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return openai_client(**request.param)


def test_chat_completion_azure_key(azure_openai_client_fixture: Callable[..., Any]) -> None:
    azure_openai_client_fixture = openai_client(api_type="azure_openai", api_key=os.environ["AZURE_OPENAI_KEY"])
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="Hello, world!"),
            UserMessage(content="What is the capital of France?"),
        ],
        max_completion_tokens=100,
    )
    response = chat_completion(request, "azure_openai", azure_openai_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


# endregion


# region Ollama


@pytest.fixture(
    params=[
        {},
    ]
)
def ollama_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return ollama_client(**request.param)


def test_chat_completion_ollama(ollama_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="llama3.2-vision:11b-instruct-q4_K_M",
        messages=[
            SystemMessage(content="Hello, world!"),
            UserMessage(content="What is the capital of France?"),
        ],
        max_completion_tokens=100,
        frequency_penalty=1.2,
        top_p=0.8,
        context_window=1000,
    )
    response = chat_completion(request, "ollama", ollama_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_ollama_structured_output(ollama_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant"),
        UserMessage(content="9.11 and 9.9 -- which is bigger?"),
    ]
    json_schema = {
        "name": "reasoning_schema",
        "strict": True,
        "schema": {
            "type": "object",
            "properties": {
                "reasoning_steps": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The reasoning steps leading to the final conclusion.",
                },
                "answer": {
                    "type": "string",
                    "description": "The final answer, taking into account the reasoning steps.",
                },
            },
            "required": ["reasoning_steps", "answer"],
            "additionalProperties": False,
        },
        "description": "A schema for structured output that includes reasoning steps and the final answer.",
    }

    request = ChatCompletionRequest(
        messages=messages,
        model="llama3.2-vision:11b-instruct-q4_K_M",
        max_completion_tokens=400,
        temperature=0.2,
        json_mode=False,
        structured_outputs=json_schema,
    )

    response = chat_completion(request, "ollama", ollama_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_ollama_multiple_tools(ollama_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(content="Call the get_current_weather function once for each city that the user mentions."),
        UserMessage(content="What's the current weather like in Boston, MA and New York, NY today?"),
    ]
    request = ChatCompletionRequest(
        model="qwen2.5:14b",
        messages=messages,
        tools=tools,
        max_completion_tokens=400,
        temperature=0,
    )
    response = chat_completion(request, "ollama", ollama_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_ollama_tool_message(ollama_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant"),
        UserMessage(content="What is the weather in Boston, MA?"),
        AssistantMessage(
            content="",
            tool_calls=[
                ToolCall(
                    id="abc123",
                    function=Function(
                        name="get_current_weather",
                        arguments={"location": "Boston, MA"},
                    ),
                )
            ],
        ),
        ToolMessage(name="abc123", content="The weather in Boston, MA is 70 degrees Fahrenheit."),
    ]
    request = ChatCompletionRequest(
        model="qwen2.5:14b",
        messages=messages,
        max_completion_tokens=300,
        temperature=0.3,
    )
    response = chat_completion(request, "ollama", ollama_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_ollama_vision(ollama_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="llama3.2-vision:11b-instruct-q4_K_M",
        max_completion_tokens=200,
        temperature=0.5,
    )

    response = chat_completion(request, "ollama", ollama_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_ollama_vision_multiple_images(ollama_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="What are the animals in the images? Reply in one word for each animal."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(dog_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]
    request = ChatCompletionRequest(
        messages=messages,
        model="llama3.2-vision:11b-instruct-q4_K_M",
        max_completion_tokens=100,
    )
    response = chat_completion(request, "ollama", ollama_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_ollama_vision_multiple_messages(ollama_client_fixture: Callable[..., Any]) -> None:
    """Test with two image messages separated by an assistant message."""
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
        AssistantMessage(content="Cat"),
        UserMessage(
            content=[
                TextContent(text="What about this animal?"),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(dog_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="llama3.2-vision:11b-instruct-q4_K_M",
        max_completion_tokens=100,
    )

    response = chat_completion(request, "ollama", ollama_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_chat_completion_ollama_max_tokens(ollama_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="llama3.2-vision:11b-instruct-q4_K_M",
        messages=[UserMessage(content="What is the capital of France?")],
        max_tokens=100,
    )
    response = chat_completion(request, "ollama", ollama_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


# endregion


# region Anthropic


@pytest.fixture(
    params=[
        {},
    ]
)
def anthropic_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return anthropic_client(**request.param)


def test_anthropic_chat_completion_simple(anthropic_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="claude-3-7-sonnet-20250219",
        messages=[
            SystemMessage(content="You are a helpful assistant."),
            UserMessage(content="What is the capital of France?"),
        ],
        max_completion_tokens=200,
    )
    response = chat_completion(request, "anthropic", anthropic_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_anthropic_chat_completion_multiple(anthropic_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="claude-3-7-sonnet-20250219",
        messages=[
            SystemMessage(content="You are a helpful assistant."),
            UserMessage(content="What is the capital of France?"),
        ],
        max_completion_tokens=200,
        temperature=1,
        top_k=20,
        top_p=0.95,
    )
    response = chat_completion(request, "anthropic", anthropic_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_anthropic_chat_completion_tool(anthropic_client_fixture: Callable[..., Any]) -> None:
    stock_tool = {
        "name": "get_stock_price",
        "description": "Get the current stock price for a given ticker symbol.",
        "input_schema": {
            "type": "object",
            "properties": {
                "ticker": {"type": "string", "description": "The stock ticker symbol, e.g. AAPL for Apple Inc."}
            },
            "required": ["ticker"],
        },
    }

    request = ChatCompletionRequest(
        model="claude-3-7-sonnet-20250219",
        messages=[
            UserMessage(content="What's the S&P 500 at today?"),
        ],
        tools=[stock_tool],
        max_tokens=300,
    )
    response = chat_completion(request, "anthropic", anthropic_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_anthropic_chat_completion_multiple_tools(anthropic_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "name": "get_weather",
            "description": "Get the current weather in a given location",
            "input_schema": {
                "type": "object",
                "properties": {
                    "location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"},
                    "unit": {
                        "type": "string",
                        "enum": ["celsius", "fahrenheit"],
                        "description": "The unit of temperature, either 'celsius' or 'fahrenheit'",
                    },
                },
                "required": ["location"],
            },
        },
        {
            "name": "get_time",
            "description": "Get the current time in a given time zone",
            "input_schema": {
                "type": "object",
                "properties": {
                    "timezone": {"type": "string", "description": "The IANA time zone name, e.g. America/Los_Angeles"}
                },
                "required": ["timezone"],
            },
        },
    ]

    request = ChatCompletionRequest(
        model="claude-3-7-sonnet-20250219",
        messages=[
            UserMessage(content="What is the weather like right now in New York? Also what time is it there?"),
        ],
        tools=tools,
        temperature=1,
        max_tokens=400,
    )
    response = chat_completion(request, "anthropic", anthropic_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_anthropic_chat_completion_tool_required_name(anthropic_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "name": "get_current_weather",
            "description": "Get the current weather",
            "input_schema": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "format": {
                        "type": "string",
                        "enum": ["celsius", "fahrenheit"],
                        "description": "The temperature unit to use. Infer this from the users location.",
                    },
                },
                "required": ["location", "format"],
            },
        }
    ]
    request = ChatCompletionRequest(
        model="claude-3-7-sonnet-20250219",
        messages=[
            UserMessage(content="What's the current weather like in Boston, MA today?"),
        ],
        tools=tools,
        tool_choice="get_current_weather",
        max_completion_tokens=300,
        temperature=0,
    )
    response = chat_completion(request, "anthropic", anthropic_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_anthropic_chat_completion_tool_required(anthropic_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "name": "get_current_weather",
            "description": "Get the current weather",
            "input_schema": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    },
                    "format": {
                        "type": "string",
                        "enum": ["celsius", "fahrenheit"],
                        "description": "The temperature unit to use. Infer this from the users location.",
                    },
                },
                "required": ["location", "format"],
            },
        }
    ]
    request = ChatCompletionRequest(
        model="claude-3-7-sonnet-20250219",
        messages=[
            UserMessage(content="What's the current weather like in Boston, MA today?"),
        ],
        tools=tools,
        tool_choice="any",
        parallel_tool_calls=False,
        max_completion_tokens=300,
        temperature=0,
    )
    response = chat_completion(request, "anthropic", anthropic_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_anthropic_chat_completion_builtin_tool(anthropic_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {"type": "text_editor_20250124", "name": "str_replace_editor"},
    ]
    request = ChatCompletionRequest(
        model="claude-3-7-sonnet-20250219",
        messages=[
            UserMessage(content="There's a syntax error in my primes.py file. Can you help me fix it?"),
        ],
        tools=tools,
        tool_choice="any",
        max_completion_tokens=500,
        temperature=0.2,
    )
    response = chat_completion(request, "anthropic", anthropic_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


def test_anthropic_chat_completion_tool_message(anthropic_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        UserMessage(content="What is the weather in Boston, MA?"),
        AssistantMessage(
            content="",
            tool_calls=[
                ToolCall(
                    id="abc123",
                    function=Function(
                        name="get_current_weather",
                        arguments={"location": "Boston, MA"},
                    ),
                )
            ],
        ),
        ToolMessage(name="abc123", content="The weather in Boston, MA is 70 degrees Fahrenheit."),
    ]
    request = ChatCompletionRequest(
        model="claude-3-7-sonnet-20250219",
        messages=messages,
        max_completion_tokens=300,
        temperature=0.3,
    )
    response = chat_completion(request, "anthropic", anthropic_client_fixture)
    print(response.model_dump(mode="json", exclude_none=True))


# endregion



import base64
from collections.abc import Sequence
from copy import deepcopy
import mimetypes
from pathlib import Path
from typing import Any

from liquid import render
from openai.lib._pydantic import to_strict_json_schema
from pydantic import BaseModel

from not_again_ai.llm.chat_completion.types import MessageT


def _apply_templates(value: Any, variables: dict[str, str]) -> Any:
    """Recursively applies Liquid templating to all string fields within the given value."""
    if isinstance(value, str):
        return render(value, **variables)
    elif isinstance(value, list):
        return [_apply_templates(item, variables) for item in value]
    elif isinstance(value, dict):
        return {key: _apply_templates(val, variables) for key, val in value.items()}
    elif isinstance(value, BaseModel):
        # Process each field in the BaseModel by converting it to a dict,
        # applying templating to its values, and then re-instantiating the model.
        processed_data = {key: _apply_templates(val, variables) for key, val in value.model_dump().items()}
        return value.__class__(**processed_data)
    else:
        return value


def compile_messages(messages: Sequence[MessageT], variables: dict[str, str]) -> Sequence[MessageT]:
    """Compiles messages using Liquid templating and the provided variables.
    Calls render(content_part, **variables) on each text content part.

    Args:
        messages: List of MessageT where content can contain Liquid templates.
        variables: The variables to inject into the templates.

    Returns:
        The same list of messages with the content parts injected with the variables.
    """
    messages_formatted = deepcopy(messages)
    messages_formatted = [_apply_templates(message, variables) for message in messages_formatted]
    return messages_formatted


def compile_tools(tools: Sequence[dict[str, Any]], variables: dict[str, str]) -> Sequence[dict[str, Any]]:
    """Compiles a list of tool argument dictionaries using Liquid templating and provided variables.

    Each dictionary in the list is deep copied and processed recursively to substitute any Liquid
    templates present in its data structure.

    Args:
        tools: A list of dictionaries representing tool arguments, where values can include Liquid templates.
        variables: A dictionary of variables to substitute into the Liquid templates.

    Returns:
        A new list of dictionaries with the Liquid templates replaced by their corresponding variable values.
    """
    tools_formatted = deepcopy(tools)
    tools_formatted = [_apply_templates(tool, variables) for tool in tools_formatted]
    return tools_formatted


def encode_image(image_path: Path) -> str:
    """Encodes an image file at the given Path to base64.

    Args:
        image_path: The path to the image file to encode.

    Returns:
        The base64 encoded image as a string.
    """
    with Path.open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def create_image_url(image_path: Path) -> str:
    """Creates a data URL for an image file at the given Path.

    Args:
        image_path: The path to the image file to encode.

    Returns:
        The data URL for the image.
    """
    image_data = encode_image(image_path)

    valid_mime_types = ["image/jpeg", "image/png", "image/webp", "image/gif"]

    # Get the MIME type from the image file extension
    mime_type = mimetypes.guess_type(image_path)[0]

    # Check if the MIME type is valid
    # List of valid types is here: https://platform.openai.com/docs/guides/vision/what-type-of-files-can-i-upload
    if mime_type not in valid_mime_types:
        raise ValueError(f"Invalid MIME type for image: {mime_type}")

    return f"data:{mime_type};base64,{image_data}"


def pydantic_to_json_schema(
    pydantic_model: type[BaseModel], schema_name: str, description: str | None = None
) -> dict[str, Any]:
    """Converts a Pydantic model to a JSON schema expected by Structured Outputs.
    Must adhere to the supported schemas: https://platform.openai.com/docs/guides/structured-outputs/supported-schemas

    Args:
        pydantic_model: The Pydantic model to convert.
        schema_name: The name of the schema.
        description: An optional description of the schema.

    Returns:
        A JSON schema dictionary representing the Pydantic model.
    """
    converted_pydantic = to_strict_json_schema(pydantic_model)
    schema = {
        "name": schema_name,
        "strict": True,
        "schema": converted_pydantic,
    }
    if description:
        schema["description"] = description
    return schema



from loguru import logger
from playwright.sync_api import Browser, Playwright, sync_playwright


def create_browser(headless: bool = True) -> tuple[Playwright, Browser]:
    """Creates and returns a new Playwright instance and browser.

    Args:
        headless (bool, optional): Whether to run the browser in headless mode. Defaults to True.

    Returns:
        tuple[Playwright, Browser]: A tuple containing the Playwright instance and browser.
    """
    pwright = sync_playwright().start()
    browser = pwright.chromium.launch(
        headless=headless,
        chromium_sandbox=False,
        timeout=15000,
    )
    return pwright, browser


def get_raw_web_content(url: str, browser: Browser | None = None, headless: bool = True) -> str:
    """Fetches raw web content from a given URL using Playwright.

    Args:
        url (str): The URL to fetch content from.
        browser (Browser | None, optional): An existing browser instance to use. Defaults to None.
        headless (bool, optional): Whether to run the browser in headless mode. Defaults to True.

    Returns:
        str: The raw web content.
    """
    p = None
    try:
        if browser is None:
            p, browser = create_browser(headless)

        page = browser.new_page(
            accept_downloads=False,
            java_script_enabled=True,
            viewport={"width": 1366, "height": 768},
            user_agent="Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.3",
        )
        page.goto(url)
        content = page.content()
        page.close()
        return content
    except Exception as e:
        logger.error(f"Failed to get web content: {e}")
        return ""
    finally:
        if browser:
            browser.close()
        if p:
            p.stop()



from collections.abc import AsyncGenerator, Callable
import json
import os
import re
import time
from typing import Any, Literal, cast

from loguru import logger
from ollama import AsyncClient, ChatResponse, Client, ResponseError

from not_again_ai.llm.chat_completion.types import (
    AssistantMessage,
    ChatCompletionChoice,
    ChatCompletionChoiceStream,
    ChatCompletionChunk,
    ChatCompletionDelta,
    ChatCompletionRequest,
    ChatCompletionResponse,
    Function,
    PartialFunction,
    PartialToolCall,
    Role,
    ToolCall,
)

OLLAMA_PARAMETER_MAP = {
    "frequency_penalty": "repeat_penalty",
    "max_completion_tokens": "num_predict",
    "context_window": "num_ctx",
    "n": None,
    "tool_choice": None,
    "reasoning_effort": None,
    "parallel_tool_calls": None,
    "logit_bias": None,
    "top_logprobs": None,
    "presence_penalty": None,
    "max_tokens": "num_predict",
}


def validate(request: ChatCompletionRequest) -> None:
    if request.json_mode and request.structured_outputs is not None:
        raise ValueError("json_schema and json_mode cannot be used together.")

    # Check if any of the parameters set to OLLAMA_PARAMETER_MAP are not None
    for key, value in OLLAMA_PARAMETER_MAP.items():
        if value is None and getattr(request, key) is not None:
            logger.warning(f"Parameter {key} is not supported by Ollama and will be ignored.")

    # If "stop" is not None, check if it is just a string
    if isinstance(request.stop, list):
        logger.warning("Parameter 'stop' needs to be a string and not a list. It will be ignored.")
        request.stop = None

    # Raise an error if both "max_tokens" and "max_completion_tokens" are provided
    if request.max_tokens is not None and request.max_completion_tokens is not None:
        raise ValueError("`max_tokens` and `max_completion_tokens` cannot both be provided.")


def format_kwargs(request: ChatCompletionRequest) -> dict[str, Any]:
    kwargs = request.model_dump(mode="json", exclude_none=True)
    # For each key in OLLAMA_PARAMETER_MAP
    # If it is not None, set the key in kwargs to the value of the corresponding value in OLLAMA_PARAMETER_MAP
    # If it is None, remove that key from kwargs
    for key, value in OLLAMA_PARAMETER_MAP.items():
        if value is not None and key in kwargs:
            kwargs[value] = kwargs.pop(key)
        elif value is None and key in kwargs:
            del kwargs[key]

    # If json_mode is True, set the format to json
    json_mode = kwargs.get("json_mode", None)
    if json_mode:
        kwargs["format"] = "json"
        kwargs.pop("json_mode")
    elif json_mode is not None and not json_mode:
        kwargs.pop("json_mode")

    # If structured_outputs is not None, set the format to structured_outputs
    if kwargs.get("structured_outputs", None):
        # Check if the schema is in the OpenAI and pull out the schema
        if "schema" in kwargs["structured_outputs"]:
            kwargs["format"] = kwargs["structured_outputs"]["schema"]
            kwargs.pop("structured_outputs")
        else:
            kwargs["format"] = kwargs.pop("structured_outputs")

    option_fields = [
        "mirostat",
        "mirostat_eta",
        "mirostat_tau",
        "num_ctx",
        "repeat_last_n",
        "repeat_penalty",
        "temperature",
        "seed",
        "stop",
        "tfs_z",
        "num_predict",
        "top_k",
        "top_p",
        "min_p",
    ]
    # For each field in option_fields, if it is in kwargs, make it under an options dictionary
    options = {}
    for field in option_fields:
        if field in kwargs:
            options[field] = kwargs.pop(field)
    kwargs["options"] = options

    for message in kwargs["messages"]:
        role = message.get("role", None)
        # For each ToolMessage, remove the name field
        if role is not None and role == "tool":
            message.pop("name")

        # For each AssistantMessage with tool calls, remove the id field
        if role is not None and role == "assistant" and message.get("tool_calls", None):
            for tool_call in message["tool_calls"]:
                tool_call.pop("id")

        # Content and images need to be separated
        images = []
        content = ""
        if isinstance(message["content"], list):
            for item in message["content"]:
                if item["type"] == "image_url":
                    image_url = item["image_url"]["url"]
                    # Remove the data URL prefix if present
                    if image_url.startswith("data:"):
                        image_url = image_url.split("base64,", 1)[1]
                    images.append(image_url)
                else:
                    content += item["text"]
        else:
            content = message["content"]

        message["content"] = content
        if len(images) > 1:
            images = images[:1]
            logger.warning("Ollama model only supports a single image per message. Using only the first images.")
        message["images"] = images

    return kwargs


def ollama_chat_completion(
    request: ChatCompletionRequest,
    client: Callable[..., Any],
) -> ChatCompletionResponse:
    validate(request)
    kwargs = format_kwargs(request)

    try:
        start_time = time.time()
        response: ChatResponse = client(**kwargs)
        end_time = time.time()
        response_duration = round(end_time - start_time, 4)
    except ResponseError as e:
        # If the error says "model 'model' not found" use regex then raise a more specific error
        expected_pattern = f"model '{request.model}' not found"
        if re.search(expected_pattern, e.error):
            raise ResponseError(f"Model '{request.model}' not found.") from e
        else:
            raise ResponseError(e.error) from e

    errors = ""

    # Handle tool calls
    tool_calls: list[ToolCall] | None = None
    if response.message.tool_calls:
        parsed_tool_calls: list[ToolCall] = []
        for tool_call in response.message.tool_calls:
            tool_name = tool_call.function.name
            if request.tools and tool_name not in [tool["function"]["name"] for tool in request.tools]:
                errors += f"Tool call {tool_call} has an invalid tool name: {tool_name}\n"
            tool_args = dict(tool_call.function.arguments)
            parsed_tool_calls.append(
                ToolCall(
                    id="",
                    function=Function(
                        name=tool_name,
                        arguments=tool_args,
                    ),
                )
            )
        tool_calls = parsed_tool_calls

    json_message = None
    if (request.json_mode or (request.structured_outputs is not None)) and response.message.content:
        try:
            json_message = json.loads(response.message.content)
        except json.JSONDecodeError:
            errors += "Message failed to parse into JSON\n"

    finish_reason = cast(
        Literal["stop", "length", "tool_calls", "content_filter"],
        "stop" if response.done_reason is None else response.done_reason or "stop",
    )

    choice = ChatCompletionChoice(
        message=AssistantMessage(
            content=response.message.content or "",
            tool_calls=tool_calls,
        ),
        finish_reason=finish_reason,
        json_message=json_message,
    )

    return ChatCompletionResponse(
        choices=[choice],
        errors=errors.strip(),
        completion_tokens=response.get("eval_count", -1),
        prompt_tokens=response.get("prompt_eval_count", -1),
        response_duration=response_duration,
    )


async def ollama_chat_completion_stream(
    request: ChatCompletionRequest,
    client: Callable[..., Any],
) -> AsyncGenerator[ChatCompletionChunk, None]:
    validate(request)
    kwargs = format_kwargs(request)

    start_time = time.time()
    stream = await client(**kwargs)

    async for chunk in stream:
        errors = ""
        # Handle tool calls
        tool_calls: list[PartialToolCall] | None = None
        if chunk.message.tool_calls:
            parsed_tool_calls: list[PartialToolCall] = []
            for tool_call in chunk.message.tool_calls:
                tool_name = tool_call.function.name
                if request.tools and tool_name not in [tool["function"]["name"] for tool in request.tools]:
                    errors += f"Tool call {tool_call} has an invalid tool name: {tool_name}\n"
                tool_args = tool_call.function.arguments

                parsed_tool_calls.append(
                    PartialToolCall(
                        id="",
                        function=PartialFunction(
                            name=tool_name,
                            arguments=tool_args,
                        ),
                    )
                )
            tool_calls = parsed_tool_calls

        current_time = time.time()
        response_duration = round(current_time - start_time, 4)

        delta = ChatCompletionDelta(
            content=chunk.message.content or "",
            role=Role.ASSISTANT,
            tool_calls=tool_calls,
        )
        choice_obj = ChatCompletionChoiceStream(
            delta=delta,
            finish_reason=chunk.done_reason,
            index=0,
        )
        chunk_obj = ChatCompletionChunk(
            choices=[choice_obj],
            errors=errors.strip(),
            completion_tokens=chunk.get("eval_count", None),
            prompt_tokens=chunk.get("prompt_eval_count", None),
            response_duration=response_duration,
        )
        yield chunk_obj


def ollama_client(
    host: str | None = None, timeout: float | None = None, async_client: bool = False
) -> Callable[..., Any]:
    """Create an Ollama client instance based on the specified host or will read from the OLLAMA_HOST environment variable.

    Args:
        host (str, optional): The host URL of the Ollama server.
        timeout (float, optional): The timeout for requests

    Returns:
        Client: An instance of the Ollama client.

    Examples:
        >>> client = client(host="http://localhost:11434")
    """
    if host is None:
        host = os.getenv("OLLAMA_HOST")
        if host is None:
            logger.warning("OLLAMA_HOST environment variable not set, using default host: http://localhost:11434")
            host = "http://localhost:11434"

    def client_callable(**kwargs: Any) -> Any:
        client = AsyncClient(host=host, timeout=timeout) if async_client else Client(host=host, timeout=timeout)
        return client.chat(**kwargs)

    return client_callable



import numpy as np

from not_again_ai.viz.distributions import univariate_distplot


def test_univariate_distplot() -> None:
    univariate_distplot(data=np.random.normal(size=1000), save_pathname=".nox/temp/distributions_test1.png")
    univariate_distplot(
        data=np.random.normal(size=1000),
        save_pathname=".nox/temp/distributions_test2.png",
        print_summary=False,
        title="Test title",
        xlabel="Test xlabel",
        ylabel="Test ylabel",
        xlim=(-5, 5),
        ylim=(0, 40),
        xticks=None,
        yticks=None,
        bins=200,
        font_size=54,
        height=14,
        aspect=2.2,
    )
    univariate_distplot(
        data=[1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10],
        save_pathname=".nox/temp/distributions_test3.png",
        print_summary=False,
        xlim=None,
        ylim=None,
        xticks=np.arange(0, 11, 1),
        yticks=np.arange(0, 5, 1),
        bins=100,
    )
    univariate_distplot(
        data=np.random.beta(a=0.5, b=0.5, size=10000),
        save_pathname=".nox/temp/distributions_test4.svg",
        print_summary=False,
        bins=100,
        title=r"Beta Distribution $\alpha=0.5, \beta=0.5$",
        font_size=18,
        height=3.91,
        aspect=1.8,
    )



from copy import deepcopy
from pathlib import Path

from pydantic import BaseModel

from not_again_ai.llm.chat_completion.types import (
    AssistantMessage,
    DeveloperMessage,
    Function,
    ImageContent,
    ImageDetail,
    ImageUrl,
    MessageT,
    SystemMessage,
    TextContent,
    ToolCall,
    UserMessage,
)
from not_again_ai.llm.prompting.compile_prompt import (
    compile_messages,
    compile_tools,
    create_image_url,
    pydantic_to_json_schema,
)

image_dir = Path(__file__).parent.parent / "sample_images"
sk_infographic = image_dir / "SKInfographic.png"
sk_diagram = image_dir / "SKDiagram.png"
cat_image = image_dir / "cat.jpg"
dog_image = image_dir / "dog.jpg"


def test_chat_prompt_no_vars() -> None:
    no_vars_prompt: list[MessageT] = [
        SystemMessage(
            content="This is a system message with no variables",
        ),
        UserMessage(
            content="This is a user message with no variables",
        ),
    ]

    messages_formatted = compile_messages(no_vars_prompt, {})
    messages_expected = [
        SystemMessage(
            content="This is a system message with no variables",
        ),
        UserMessage(
            content="This is a user message with no variables",
        ),
    ]
    assert messages_formatted == messages_expected


def test_compile_messages_place_extraction() -> None:
    place_extraction_prompt: list[MessageT] = [
        SystemMessage(
            content="""- You are a helpful assistant trying to extract places that occur in a given text.
- You must identify all the places in the text and return them in a list like this: ["place1", "place2", "place3"].""",
        ),
        UserMessage(
            content="""Here is the text I want you to extract places from:
{%- # The user's input text goes below %}
{{text}}""",
        ),
    ]
    variables = {
        "text": "I went to Paris and Berlin.",
    }
    messages_formatted = compile_messages(place_extraction_prompt, variables)
    messages_expected = [
        SystemMessage(
            content="""- You are a helpful assistant trying to extract places that occur in a given text.
- You must identify all the places in the text and return them in a list like this: ["place1", "place2", "place3"].""",
        ),
        UserMessage(
            content="""Here is the text I want you to extract places from:
I went to Paris and Berlin.""",
        ),
    ]
    assert messages_formatted == messages_expected


def test_compile_messages_vision() -> None:
    vision_prompt: list[MessageT] = [
        SystemMessage(
            content="You are a helpful assistant.",
        ),
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(image_url=ImageUrl(url=create_image_url(cat_image), detail=ImageDetail.LOW)),
            ],
        ),
        AssistantMessage(
            content="{{ answer }}",
        ),
        UserMessage(
            content=[
                TextContent(text="What about this animal?"),
                ImageContent(image_url=ImageUrl(url=create_image_url(dog_image), detail=ImageDetail.LOW)),
            ],
        ),
    ]

    messages_formatted = compile_messages(vision_prompt, {"answer": "Cat"})
    messages_expected = [
        SystemMessage(
            content="You are a helpful assistant.",
        ),
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(image_url=ImageUrl(url=create_image_url(cat_image), detail=ImageDetail.LOW)),
            ],
        ),
        AssistantMessage(
            content="Cat",
        ),
        UserMessage(
            content=[
                TextContent(text="What about this animal?"),
                ImageContent(image_url=ImageUrl(url=create_image_url(dog_image), detail=ImageDetail.LOW)),
            ],
        ),
    ]

    assert messages_formatted == messages_expected


def test_compile_messages_templates_other_string_fields() -> None:
    # Variables for Liquid template replacement.
    variables = {
        "user": "Alice",
        "func_var": "add",
        "toolcall_id": "toolcall1",
    }

    # Create messages with Liquid templates in 'name' fields and within a Function.
    system_message = SystemMessage(content="System content", name="System-{{user}}")
    developer_message = DeveloperMessage(content="Developer message", name="Dev-{{user}}")
    assistant_message = AssistantMessage(
        content="Assistant processing {{user}}",
        name="Assistant-{{user}}",
        tool_calls=[
            ToolCall(
                id="{{toolcall_id}}",
                function=Function(
                    name="function-{{func_var}}",
                    arguments={"op": "{{func_var}}", "constant": "nochange"},
                ),
            )
        ],
    )

    messages: list[MessageT] = [system_message, developer_message, assistant_message]
    compiled_messages = compile_messages(messages, variables)

    # The expected messages assume that Liquid templates in all string fields are processed.
    expected_system_message = SystemMessage(content="System content", name="System-Alice")
    expected_developer_message = DeveloperMessage(content="Developer message", name="Dev-Alice")
    expected_assistant_message = AssistantMessage(
        content="Assistant processing Alice",
        name="Assistant-Alice",
        tool_calls=[
            ToolCall(
                id="toolcall1",
                function=Function(
                    name="function-add",
                    arguments={"op": "add", "constant": "nochange"},
                ),
            )
        ],
    )
    expected_messages = [
        expected_system_message,
        expected_developer_message,
        expected_assistant_message,
    ]

    assert compiled_messages == expected_messages


def test_compile_messages_idempotence() -> None:
    """Test that compiling messages twice yields the same result."""
    msg = SystemMessage(content="Hello {{name}}", name="System-{{name}}")
    first_compile = compile_messages([msg], {"name": "World"})[0]
    second_compile = compile_messages([first_compile], {"name": "World"})[0]
    assert first_compile == second_compile


def test_compile_messages_does_not_modify_original() -> None:
    """Test that compile_messages does not modify the original messages."""
    original_msg = [SystemMessage(content="Hello {{name}}", name="System{{name}}")]
    messages_copy = deepcopy(original_msg)
    compiled_messages = compile_messages(original_msg, {"name": "Test"})
    # The original message should remain unchanged.
    assert original_msg == messages_copy
    # The compiled message should have substitutions applied.
    assert compiled_messages[0].content == "Hello Test"
    assert compiled_messages[0].name == "SystemTest"


def test_nested_structure_in_tool_calls() -> None:
    """Test that templating is recursively applied within nested structures such as tool_calls."""
    msg = AssistantMessage(
        content="Processing {{value}}",
        tool_calls=[
            ToolCall(
                id="{{call_id}}",
                function=Function(
                    name="func{{value}}",
                    arguments={"nested": {"key": "{{value}}"}},
                ),
            )
        ],
    )
    compiled_messages = compile_messages([msg], {"value": "100", "call_id": "call_1"})
    assert isinstance(compiled_messages[0], AssistantMessage), "Expected an AssistantMessage"
    compiled: AssistantMessage = compiled_messages[0]
    assert compiled.content == "Processing 100"
    assert compiled.tool_calls is not None
    tool_call = compiled.tool_calls[0]
    assert tool_call.id == "call_1"
    assert tool_call.function.name == "func100"
    assert tool_call.function.arguments == {"nested": {"key": "100"}}


def test_missing_variable_in_template() -> None:
    """Test that a missing variable in the template yields an empty substitution."""
    msg = SystemMessage(content="Hello {{name}}")
    compiled = compile_messages([msg], {})[0]
    # Assuming the missing variable renders as an empty string.
    assert compiled.content == "Hello "


def test_compile_tools_basic() -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. {{location}}",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use, such as {{format}}.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        }
    ]

    expected_tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use, such as celsius.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        }
    ]

    compiled_tools = compile_tools(tools, {"location": "San Francisco, CA", "format": "celsius"})
    assert compiled_tools == expected_tools


def test_compile_tools_no_change() -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state",
                        },
                    },
                    "required": ["location"],
                },
            },
        }
    ]
    assert compile_tools(tools, {}) == tools


def test_compile_tools_does_not_modify_original() -> None:
    """Test that compile_tools does not mutate the original list of tool dictionaries."""
    original_tools = [{"name": "Template: {{var}}"}]
    tools_copy = deepcopy(original_tools)
    variables = {"var": "A"}
    result = compile_tools(original_tools, variables)
    # Confirm that the original remains unchanged.
    assert original_tools == tools_copy
    # And that the result is modified as expected.
    expected = [{"name": "Template: A"}]
    assert result == expected


def test_pydantic_to_json_schema() -> None:
    class Step(BaseModel):
        explanation: str
        output: str

    class MathResponse(BaseModel):
        steps: list[Step]
        final_answer: str

    json_schema = pydantic_to_json_schema(MathResponse, schema_name="math_response", description="A math response")

    expected_schema = {
        "$defs": {
            "Step": {
                "properties": {
                    "explanation": {"title": "Explanation", "type": "string"},
                    "output": {"title": "Output", "type": "string"},
                },
                "required": ["explanation", "output"],
                "title": "Step",
                "type": "object",
                "additionalProperties": False,
            }
        },
        "properties": {
            "steps": {"items": {"$ref": "#/$defs/Step"}, "title": "Steps", "type": "array"},
            "final_answer": {"title": "Final Answer", "type": "string"},
        },
        "required": ["steps", "final_answer"],
        "title": "MathResponse",
        "type": "object",
        "additionalProperties": False,
    }

    assert json_schema["strict"] is True
    assert json_schema["name"] == "math_response"
    assert json_schema["description"] == "A math response"
    assert json_schema["schema"] == expected_schema



from not_again_ai.viz.barplots import simple_barplot


def test_simple_barplot() -> None:
    save_pathname = ".nox/temp/barplot_test1.png"
    x = ["fence", "wall", "gate", "door", "window", "counter", "stair", "curtain", "ceiling", "floor"]
    y = [0.5, 0.01, 0.25, 0.2, 0.1, 0.05, 0.04, 0.03, 0.02, 0.3]
    simple_barplot(
        x,
        y,
        save_pathname,
        order="asc",
        title="Token logits",
        x_label="token",
        y_label="logit",
    )

    save_pathname = ".nox/temp/barplot_test2.png"
    simple_barplot(
        y,
        x,
        save_pathname,
        order="desc",
        orient_bars_vertically=False,
        title="Token logits",
        x_label="logit",
        y_label="token",
    )

    save_pathname = ".nox/temp/barplot_test3.png"
    simple_barplot(
        y,
        x,
        save_pathname,
        order="asc",
        orient_bars_vertically=False,
        title="Token logits",
        x_label="logit",
        y_label="token",
    )

    save_pathname = ".nox/temp/barplot_test4.png"
    simple_barplot(
        x,
        y,
        save_pathname,
        order="desc",
        orient_bars_vertically=True,
        title="Token logits",
        x_label="logit",
        y_label="token",
    )



import numpy as np

from not_again_ai.viz.scatterplot import scatterplot_basic


def test_scatterplot_basic() -> None:
    rs = np.random.RandomState(365)
    x = rs.randn(100)
    e = rs.randn(100) * 0.2
    y = x + e
    scatterplot_basic(x, y, save_pathname=".nox/temp/scatterplot_basic1.png", title="Correlation Chart")
    scatterplot_basic(
        x,
        y,
        save_pathname=".nox/temp/scatterplot_basic2.png",
        title=None,
        xlim=(-10, 10),
        ylim=(-5, 5),
        font_size=36,
        height=15,
        aspect=2.2,
    )



from collections.abc import AsyncGenerator, Callable, Coroutine
import json
import time
from typing import Any, Literal

from azure.identity import DefaultAzureCredential, get_bearer_token_provider
from openai import AsyncAzureOpenAI, AsyncOpenAI, AzureOpenAI, OpenAI

from not_again_ai.llm.chat_completion.types import (
    AssistantMessage,
    ChatCompletionChoice,
    ChatCompletionChoiceStream,
    ChatCompletionChunk,
    ChatCompletionDelta,
    ChatCompletionRequest,
    ChatCompletionResponse,
    Function,
    PartialFunction,
    PartialToolCall,
    Role,
    ToolCall,
)

OPENAI_PARAMETER_MAP = {
    "context_window": None,
    "mirostat": None,
    "mirostat_eta": None,
    "mirostat_tau": None,
    "repeat_last_n": None,
    "tfs_z": None,
    "top_k": None,
    "min_p": None,
}


def validate(request: ChatCompletionRequest) -> None:
    if request.json_mode and request.structured_outputs is not None:
        raise ValueError("json_schema and json_mode cannot be used together.")

    # Raise an error if both "max_tokens" and "max_completion_tokens" are provided
    if request.max_tokens is not None and request.max_completion_tokens is not None:
        raise ValueError("`max_tokens` and `max_completion_tokens` cannot both be provided.")


def format_kwargs(request: ChatCompletionRequest) -> dict[str, Any]:
    # Format the response format parameters to be compatible with OpenAI API
    if request.json_mode:
        response_format: dict[str, Any] = {"type": "json_object"}
    elif request.structured_outputs is not None:
        response_format = {"type": "json_schema", "json_schema": request.structured_outputs}
    else:
        response_format = {"type": "text"}

    kwargs = request.model_dump(mode="json", exclude_none=True)

    # For each key in OPENAI_PARAMETER_MAP
    # If it is not None, set the key in kwargs to the value of the corresponding value in OPENAI_PARAMETER_MAP
    # If it is None, remove that key from kwargs
    for key, value in OPENAI_PARAMETER_MAP.items():
        if value is not None and key in kwargs:
            kwargs[value] = kwargs.pop(key)
        elif value is None and key in kwargs:
            del kwargs[key]

    for message in kwargs["messages"]:
        role = message.get("role", None)
        # For each ToolMessage, change the "name" field to be named "tool_call_id" instead
        if role is not None and role == "tool":
            message["tool_call_id"] = message.pop("name")

        # For each AssistantMessage with tool calls, make the function arguments a string
        if role is not None and role == "assistant" and message.get("tool_calls", None):
            for tool_call in message["tool_calls"]:
                tool_call["function"]["arguments"] = str(tool_call["function"]["arguments"])

    # Delete the json_mode and structured_outputs from kwargs
    kwargs.pop("json_mode", None)
    kwargs.pop("structured_outputs", None)

    # Add the response_format to kwargs
    kwargs["response_format"] = response_format

    # Handle tool_choice when the provided tool_choice the name of the required tool.
    if request.tool_choice is not None and request.tool_choice not in ["none", "auto", "required"]:
        kwargs["tool_choice"] = {"type": "function", "function": {"name": request.tool_choice}}

    return kwargs


def process_logprobs(logprobs_content: list[dict[str, Any]]) -> list[dict[str, Any] | list[dict[str, Any]]]:
    """Process logprobs content from OpenAI API response.

    Args:
        logprobs_content: List of logprob entries from the API response

    Returns:
        Processed logprobs list containing either single token info or lists of top token infos
    """
    logprobs_list: list[dict[str, Any] | list[dict[str, Any]]] = []
    for logprob in logprobs_content:
        if logprob.get("top_logprobs", None):
            curr_logprob_infos: list[dict[str, Any]] = []
            for top_logprob in logprob.get("top_logprobs", []):
                curr_logprob_infos.append(
                    {
                        "token": top_logprob.get("token", ""),
                        "logprob": top_logprob.get("logprob", 0),
                        "bytes": top_logprob.get("bytes", 0),
                    }
                )
            logprobs_list.append(curr_logprob_infos)
        else:
            logprobs_list.append(
                {
                    "token": logprob.get("token", ""),
                    "logprob": logprob.get("logprob", 0),
                    "bytes": logprob.get("bytes", 0),
                }
            )
    return logprobs_list


def openai_chat_completion(
    request: ChatCompletionRequest,
    client: Callable[..., Any],
) -> ChatCompletionResponse:
    validate(request)
    kwargs = format_kwargs(request)

    start_time = time.time()
    response = client(**kwargs)
    end_time = time.time()
    response_duration = round(end_time - start_time, 4)

    errors = ""
    extras: dict[str, Any] = {}
    choices: list[ChatCompletionChoice] = []
    for index, choice in enumerate(response["choices"]):
        choice_extras: dict[str, Any] = {}
        finish_reason = choice["finish_reason"]

        message = choice["message"]
        tool_calls: list[ToolCall] | None = None
        if message.get("tool_calls", None):
            parsed_tool_calls: list[ToolCall] = []
            for tool_call in message["tool_calls"]:
                tool_name = tool_call.get("function", {}).get("name", None)
                # Check if the tool name is valid (one of the tool names in the request)
                if request.tools and tool_name not in [tool["function"]["name"] for tool in request.tools]:
                    errors += f"Choice {index}: Tool call {tool_call} has an invalid tool name: {tool_name}\n"

                tool_args = tool_call.get("function", {}).get("arguments", None)
                try:
                    tool_args = json.loads(tool_args)
                except json.JSONDecodeError:
                    errors += f"Choice {index}: Tool call {tool_call} failed to parse arguments into JSON\n"

                parsed_tool_calls.append(
                    ToolCall(
                        id=tool_call["id"],
                        function=Function(
                            name=tool_name,
                            arguments=tool_args,
                        ),
                    )
                )
            tool_calls = parsed_tool_calls

        json_message = None
        if request.json_mode or (request.structured_outputs is not None):
            try:
                json_message = json.loads(message.get("content", "{}"))
            except json.JSONDecodeError:
                errors += f"Choice {index}: Message failed to parse into JSON\n"

        # Handle logprobs
        logprobs: list[dict[str, Any] | list[dict[str, Any]]] | None = None
        if choice.get("logprobs", None) and choice["logprobs"].get("content", None) is not None:
            logprobs = process_logprobs(choice["logprobs"]["content"])

        # Handle extras that OpenAI or Azure OpenAI return
        if choice.get("content_filter_results", None):
            choice_extras["content_filter_results"] = choice["content_filter_results"]

        choices.append(
            ChatCompletionChoice(
                message=AssistantMessage(
                    content=message.get("content") or "",
                    refusal=message.get("refusal", None),
                    tool_calls=tool_calls,
                ),
                finish_reason=finish_reason,
                json_message=json_message,
                logprobs=logprobs,
                extras=choice_extras,
            )
        )

    completion_tokens = response["usage"].get("completion_tokens", -1)
    prompt_tokens = response["usage"].get("prompt_tokens", -1)
    completion_detailed_tokens = response["usage"].get("completion_detailed_tokens", None)
    prompt_detailed_tokens = response["usage"].get("prompt_detailed_tokens", None)
    system_fingerprint = response.get("system_fingerprint", None)

    extras["prompt_filter_results"] = response.get("prompt_filter_results", None)

    return ChatCompletionResponse(
        choices=choices,
        errors=errors.strip(),
        extras=extras,
        completion_tokens=completion_tokens,
        prompt_tokens=prompt_tokens,
        completion_detailed_tokens=completion_detailed_tokens,
        prompt_detailed_tokens=prompt_detailed_tokens,
        system_fingerprint=system_fingerprint,
        response_duration=response_duration,
    )


async def openai_chat_completion_stream(
    request: ChatCompletionRequest,
    client: Callable[..., Any],
) -> AsyncGenerator[ChatCompletionChunk, None]:
    validate(request)
    kwargs = format_kwargs(request)

    start_time = time.time()
    stream = await client(**kwargs)

    async for chunk in stream:
        errors = ""
        # This kind of a hack. To make this processing generic for clients that do not return the correct
        # data structure, we convert the chunk to a dict
        if not isinstance(chunk, dict):
            chunk = chunk.to_dict()

        choices: list[ChatCompletionChoiceStream] = []
        for choice in chunk["choices"]:
            content = choice.get("delta", {}).get("content", "")
            if not content:
                content = ""

            role = Role.ASSISTANT
            if choice.get("delta", {}).get("role", None):
                role = Role(choice["delta"]["role"])

            # Handle tool calls
            tool_calls: list[PartialToolCall] | None = None
            if choice["delta"].get("tool_calls", None):
                parsed_tool_calls: list[PartialToolCall] = []
                for tool_call in choice["delta"]["tool_calls"]:
                    tool_name = tool_call.get("function", {}).get("name", None)
                    if not tool_name:
                        tool_name = ""
                    tool_args = tool_call.get("function", {}).get("arguments", "")
                    if not tool_args:
                        tool_args = ""

                    tool_id = tool_call.get("id", None)
                    parsed_tool_calls.append(
                        PartialToolCall(
                            id=tool_id,
                            function=PartialFunction(
                                name=tool_name,
                                arguments=tool_args,
                            ),
                        )
                    )
                tool_calls = parsed_tool_calls

            refusal = None
            if choice["delta"].get("refusal", None):
                refusal = choice["delta"]["refusal"]

            delta = ChatCompletionDelta(
                content=content,
                role=role,
                tool_calls=tool_calls,
                refusal=refusal,
            )

            index = choice.get("index", 0)
            finish_reason = choice.get("finish_reason", None)

            # Handle logprobs
            logprobs: list[dict[str, Any] | list[dict[str, Any]]] | None = None
            if choice.get("logprobs", None) and choice["logprobs"].get("content", None) is not None:
                logprobs = process_logprobs(choice["logprobs"]["content"])

            choice_obj = ChatCompletionChoiceStream(
                delta=delta,
                finish_reason=finish_reason,
                logprobs=logprobs,
                index=index,
            )
            choices.append(choice_obj)

        current_time = time.time()
        response_duration = round(current_time - start_time, 4)

        if "usage" in chunk and chunk["usage"] is not None:
            completion_tokens = chunk["usage"].get("completion_tokens", None)
            prompt_tokens = chunk["usage"].get("prompt_tokens", None)
            system_fingerprint = chunk.get("system_fingerprint", None)
        else:
            completion_tokens = None
            prompt_tokens = None
            system_fingerprint = None

        chunk_obj = ChatCompletionChunk(
            choices=choices,
            errors=errors.strip(),
            completion_tokens=completion_tokens,
            prompt_tokens=prompt_tokens,
            response_duration=response_duration,
            system_fingerprint=system_fingerprint,
        )
        yield chunk_obj


def create_client_callable(client_class: type[OpenAI | AzureOpenAI], **client_args: Any) -> Callable[..., Any]:
    """Creates a callable that instantiates and uses an OpenAI client.

    Args:
        client_class: The OpenAI client class to instantiate (OpenAI or AzureOpenAI)
        **client_args: Arguments to pass to the client constructor

    Returns:
        A callable that creates a client and returns completion results
    """
    filtered_args = {k: v for k, v in client_args.items() if v is not None}

    def client_callable(**kwargs: Any) -> Any:
        client = client_class(**filtered_args)
        completion = client.chat.completions.create(**kwargs)
        return completion.to_dict()

    return client_callable


def create_client_callable_stream(
    client_class: type[AsyncOpenAI | AsyncAzureOpenAI], **client_args: Any
) -> Callable[..., Any]:
    filtered_args = {k: v for k, v in client_args.items() if v is not None}

    def client_callable(**kwargs: Any) -> Coroutine[Any, Any, Any]:
        client = client_class(**filtered_args)
        kwargs["stream_options"] = {"include_usage": True}
        stream = client.chat.completions.create(**kwargs)
        return stream

    return client_callable


class InvalidOAIAPITypeError(Exception):
    """Raised when an invalid OAIAPIType string is provided."""


def openai_client(
    api_type: Literal["openai", "azure_openai"] = "openai",
    api_key: str | None = None,
    organization: str | None = None,
    aoai_api_version: str = "2024-06-01",
    azure_endpoint: str | None = None,
    timeout: float | None = None,
    max_retries: int | None = None,
    async_client: bool = False,
) -> Callable[..., Any]:
    """Create an OpenAI or Azure OpenAI client instance based on the specified API type and other provided parameters.

    It is preferred to use RBAC authentication for Azure OpenAI. You must be signed in with the Azure CLI and have correct role assigned.
    See https://techcommunity.microsoft.com/t5/microsoft-developer-community/using-keyless-authentication-with-azure-openai/ba-p/4111521

    Args:
        api_type (str, optional): Type of the API to be used. Accepted values are 'openai' or 'azure_openai'.
            Defaults to 'openai'.
        api_key (str, optional): The API key to authenticate the client. If not provided,
            OpenAI automatically uses `OPENAI_API_KEY` from the environment.
            If provided for Azure OpenAI, it will be used for authentication instead of the Azure AD token provider.
        organization (str, optional): The ID of the organization. If not provided,
            OpenAI automotically uses `OPENAI_ORG_ID` from the environment.
        aoai_api_version (str, optional): Only applicable if using Azure OpenAI https://learn.microsoft.com/azure/ai-services/openai/reference#rest-api-versioning
        azure_endpoint (str, optional): The endpoint to use for Azure OpenAI.
        timeout (float, optional): By default requests time out after 10 minutes.
        max_retries (int, optional): Certain errors are automatically retried 2 times by default,
            with a short exponential backoff. Connection errors (for example, due to a network connectivity problem),
            408 Request Timeout, 409 Conflict, 429 Rate Limit, and >=500 Internal errors are all retried by default.
        async_client (bool, optional): Whether to return an async client. Defaults to False.

    Returns:
        Callable[..., Any]: A callable that creates a client and returns completion results

    Raises:
        InvalidOAIAPITypeError: If an invalid API type string is provided.
        NotImplementedError: If the specified API type is recognized but not yet supported (e.g., 'azure_openai').
    """
    if api_type not in ["openai", "azure_openai"]:
        raise InvalidOAIAPITypeError(f"Invalid OAIAPIType: {api_type}. Must be 'openai' or 'azure_openai'.")

    if api_type == "openai":
        client_class = AsyncOpenAI if async_client else OpenAI
        callable_creator = create_client_callable_stream if async_client else create_client_callable
        return callable_creator(
            client_class,  # type: ignore
            api_key=api_key,
            organization=organization,
            timeout=timeout,
            max_retries=max_retries,
        )
    elif api_type == "azure_openai":
        azure_client_class = AsyncAzureOpenAI if async_client else AzureOpenAI
        callable_creator = create_client_callable_stream if async_client else create_client_callable
        if api_key:
            return callable_creator(
                azure_client_class,  # type: ignore
                api_version=aoai_api_version,
                azure_endpoint=azure_endpoint,
                api_key=api_key,
                timeout=timeout,
                max_retries=max_retries,
            )
        else:
            azure_credential = DefaultAzureCredential()
            ad_token_provider = get_bearer_token_provider(
                azure_credential, "https://cognitiveservices.azure.com/.default"
            )
            return callable_creator(
                azure_client_class,  # type: ignore
                api_version=aoai_api_version,
                azure_endpoint=azure_endpoint,
                azure_ad_token_provider=ad_token_provider,
                timeout=timeout,
                max_retries=max_retries,
            )
    else:
        raise NotImplementedError(f"API type '{api_type}' is invalid.")



from typing import Any

from pydantic import BaseModel, Field


class EmbeddingRequest(BaseModel):
    input: str | list[str]
    model: str
    dimensions: int | None = Field(default=None)


class EmbeddingObject(BaseModel):
    embedding: list[float]
    index: int


class EmbeddingResponse(BaseModel):
    embeddings: list[EmbeddingObject]
    total_tokens: int | None = Field(default=None)
    response_duration: float

    errors: str = Field(default="")
    extras: Any | None = Field(default=None)



import matplotlib
import numpy as np
import pandas as pd

from not_again_ai.viz.time_series import ts_lineplot


def test_ts_lineplot() -> None:
    rs = np.random.RandomState(365)
    values = rs.randn(365, 4).cumsum(axis=0).T
    dates = pd.date_range("1 1 2021", periods=365, freq="D")
    ts_lineplot(
        ts_data=values,
        save_pathname=".nox/temp/ts_lineplot1.png",
        ts_x=None,
        ts_names=None,
        legend_title="Test Title",
    )
    ts_lineplot(
        ts_data=values,
        save_pathname=".nox/temp/ts_lineplot2.png",
        ts_x=dates,
        ts_names=None,
        xlabel="Months",
        ylabel="Number",
        ylim=(-15, 15),
    )
    ts_lineplot(
        ts_data=values,
        save_pathname=".nox/temp/ts_lineplot3.png",
        ts_x=None,
        ts_names=["A", "B", "C", "D"],
        height=14,
        aspect=1.5,
    )
    ts_lineplot(
        ts_data=values,
        save_pathname=".nox/temp/ts_lineplot4.png",
        ts_x=dates,
        ts_names=["A", "B", "C", "D"],
        title="Example Time Series",
        xlabel=None,
        ylabel=None,
        yticks=np.arange(-30, 40, 10),
        xaxis_date_format="%b '%y",
        xaxis_major_locator=matplotlib.dates.MonthLocator((1, 3, 5, 7, 9, 11)),
        font_size=46,
        linewidth=1.8,
        legend_loc=2,
        palette="colorblind",
    )
    ts_lineplot(
        ts_data=values,
        save_pathname=".nox/temp/ts_lineplot5.svg",
        ts_x=dates,
        ts_names=["A", "B", "C", "D"],
        title="Example Time Series",
        xlabel=None,
        ylabel=None,
        yticks=np.arange(-30, 40, 10),
        xaxis_date_format="%b '%y",
        xaxis_major_locator=matplotlib.dates.MonthLocator((1, 3, 5, 7, 9, 11)),
        linewidth=1.5,
        legend_loc=2,
        palette="colorblind",
        font_size=20,
        height=4.4,
        aspect=1.8,
    )



from enum import Enum
from typing import Any, Generic, Literal, TypeVar

from pydantic import BaseModel, Field


class Role(str, Enum):
    ASSISTANT = "assistant"
    DEVELOPER = "developer"
    SYSTEM = "system"
    TOOL = "tool"
    USER = "user"


class ContentPartType(str, Enum):
    TEXT = "text"
    IMAGE = "image_url"


class TextContent(BaseModel):
    type: Literal[ContentPartType.TEXT] = ContentPartType.TEXT
    text: str


class ImageDetail(str, Enum):
    AUTO = "auto"
    LOW = "low"
    HIGH = "high"


class ImageUrl(BaseModel):
    url: str
    detail: ImageDetail = ImageDetail.AUTO


class ImageContent(BaseModel):
    type: Literal[ContentPartType.IMAGE] = ContentPartType.IMAGE
    image_url: ImageUrl


ContentT = TypeVar("ContentT", bound=str | list[TextContent | ImageContent])


class BaseMessage(BaseModel, Generic[ContentT]):
    content: ContentT
    role: Role
    name: str | None = None


class Function(BaseModel):
    name: str
    arguments: dict[str, Any]


class PartialFunction(BaseModel):
    name: str
    arguments: str | dict[str, Any]


class ToolCall(BaseModel):
    id: str
    function: Function
    type: Literal["function"] = "function"


class PartialToolCall(BaseModel):
    id: str | None
    function: PartialFunction
    type: Literal["function"] = "function"


class DeveloperMessage(BaseMessage[str]):
    role: Literal[Role.DEVELOPER] = Role.DEVELOPER


class SystemMessage(BaseMessage[str]):
    role: Literal[Role.SYSTEM] = Role.SYSTEM


class UserMessage(BaseMessage[str | list[TextContent | ImageContent]]):
    role: Literal[Role.USER] = Role.USER


class AssistantMessage(BaseMessage[str]):
    role: Literal[Role.ASSISTANT] = Role.ASSISTANT
    refusal: str | None = None
    tool_calls: list[ToolCall] | None = None


class ToolMessage(BaseMessage[str]):
    # A tool message's name field will be interpreted as "tool_call_id"
    role: Literal[Role.TOOL] = Role.TOOL


MessageT = AssistantMessage | DeveloperMessage | SystemMessage | ToolMessage | UserMessage


class ChatCompletionRequest(BaseModel):
    messages: list[MessageT]
    model: str
    stream: bool = Field(default=False)

    max_completion_tokens: int | None = Field(default=None)
    context_window: int | None = Field(default=None)
    logprobs: bool | None = Field(default=None)
    n: int | None = Field(default=None)

    tools: list[dict[str, Any]] | None = Field(default=None)
    tool_choice: str | None = Field(default=None)
    parallel_tool_calls: bool | None = Field(default=None)
    json_mode: bool | None = Field(default=None)
    structured_outputs: dict[str, Any] | None = Field(default=None)

    temperature: float | None = Field(default=None)
    reasoning_effort: Literal["low", "medium", "high"] | None = Field(default=None)
    top_p: float | None = Field(default=None)
    logit_bias: dict[str, float] | None = Field(default=None)
    top_logprobs: int | None = Field(default=None)
    frequency_penalty: float | None = Field(default=None)
    presence_penalty: float | None = Field(default=None)
    stop: str | list[str] | None = Field(default=None)

    seed: int | None = Field(default=None)

    mirostat: int | None = Field(default=None)
    mirostat_eta: float | None = Field(default=None)
    mirostat_tau: float | None = Field(default=None)
    repeat_last_n: int | None = Field(default=None)
    tfs_z: float | None = Field(default=None)
    top_k: int | None = Field(default=None)
    min_p: float | None = Field(default=None)

    max_tokens: int | None = Field(
        default=None,
        description="Sometimes `max_completion_tokens` is not correctly supported so we provide this as a fallback.",
    )


class ChatCompletionChoice(BaseModel):
    message: AssistantMessage
    finish_reason: Literal[
        "stop", "length", "tool_calls", "content_filter", "end_turn", "max_tokens", "stop_sequence", "tool_use"
    ]

    json_message: dict[str, Any] | None = Field(default=None)
    logprobs: list[dict[str, Any] | list[dict[str, Any]]] | None = Field(default=None)

    extras: Any | None = Field(default=None)


class ChatCompletionResponse(BaseModel):
    choices: list[ChatCompletionChoice]

    errors: str = Field(default="")

    completion_tokens: int
    prompt_tokens: int
    completion_detailed_tokens: dict[str, int] | None = Field(default=None)
    prompt_detailed_tokens: dict[str, int] | None = Field(default=None)
    cache_read_input_tokens: int | None = Field(default=None)
    cache_creation_input_tokens: int | None = Field(default=None)
    response_duration: float

    system_fingerprint: str | None = Field(default=None)

    extras: Any | None = Field(default=None)


class ChatCompletionDelta(BaseModel):
    content: str
    role: Role = Field(default=Role.ASSISTANT)

    tool_calls: list[PartialToolCall] | None = Field(default=None)

    refusal: str | None = Field(default=None)


class ChatCompletionChoiceStream(BaseModel):
    delta: ChatCompletionDelta
    index: int
    finish_reason: Literal["stop", "length", "tool_calls", "content_filter"] | None

    logprobs: list[dict[str, Any] | list[dict[str, Any]]] | None = Field(default=None)

    extras: Any | None = Field(default=None)


class ChatCompletionChunk(BaseModel):
    choices: list[ChatCompletionChoiceStream]

    errors: str = Field(default="")

    completion_tokens: int | None = Field(default=None)
    prompt_tokens: int | None = Field(default=None)
    response_duration: float | None = Field(default=None)

    system_fingerprint: str | None = Field(default=None)
    extras: Any | None = Field(default=None)



from collections.abc import Callable
from pathlib import Path
from typing import Any

import pytest

from not_again_ai.llm.chat_completion import chat_completion_stream
from not_again_ai.llm.chat_completion.providers.ollama_api import ollama_client
from not_again_ai.llm.chat_completion.providers.openai_api import openai_client
from not_again_ai.llm.chat_completion.types import (
    AssistantMessage,
    ChatCompletionRequest,
    Function,
    ImageContent,
    ImageDetail,
    ImageUrl,
    MessageT,
    SystemMessage,
    TextContent,
    ToolCall,
    ToolMessage,
    UserMessage,
)
from not_again_ai.llm.prompting.compile_prompt import encode_image

image_dir = Path(__file__).parent.parent / "sample_images"
cat_image = image_dir / "cat.jpg"
dog_image = image_dir / "dog.jpg"
numbers_image = image_dir / "numbers.png"
sk_infographic = image_dir / "SKInfographic.png"
sk_diagram = image_dir / "SKDiagram.png"


# region Azure OpenAI


@pytest.fixture(
    params=[
        {"async_client": True},
        {"api_type": "azure_openai", "aoai_api_version": "2025-01-01-preview", "async_client": True},
    ]
)
def openai_aoai_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return openai_client(**request.param)


async def test_chat_completion_stream_simple(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="Hello, world!"),
            UserMessage(content="What is the capital of France?"),
        ],
        max_completion_tokens=100,
    )
    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_json_mode(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(
            content="""You are getting names of users and formatting them into json.
Example:
User: Jane Doe
Output: {"name": "Jane Doe"}"""
        ),
        UserMessage(content="John Doe"),
    ]
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=messages,
        max_completion_tokens=200,
        temperature=0,
        json_mode=True,
    )
    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_n(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="You are a helpful assistant."),
            UserMessage(content="Hello!"),
        ],
        max_completion_tokens=100,
        n=2,
    )
    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_toplogprobs(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            SystemMessage(content="You are a helpful assistant."),
            UserMessage(content="Hello!"),
        ],
        max_completion_tokens=100,
        logprobs=True,
        top_logprobs=3,
    )
    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_tool_simple(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        }
    ]

    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            UserMessage(
                content="What's the current weather like in Boston, MA today? Call the get_current_weather function."
            )
        ],
        tools=tools,
        max_completion_tokens=300,
        temperature=0,
    )

    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_tool_required_name(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[
            UserMessage(content="What's the current weather like in Boston, MA today?"),
        ],
        tools=tools,
        tool_choice="get_current_weather",
        max_completion_tokens=300,
        temperature=0,
    )
    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_multiple_tools(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(content="Call the get_current_weather function once for each city that the user mentions."),
        UserMessage(content="What's the current weather like in Boston, MA and New York, NY today?"),
    ]
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=messages,
        tools=tools,
        max_completion_tokens=400,
        temperature=0,
    )
    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_message_with_tools(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(
            content="""You will be given a function called get_current_weather.
Before calling the function, first reason about which city the user is asking about. YOU MUST think step by step before calling the function.
For example, if the user asks 'What's the current weather like in Boston, MA today?', You should first say 'The user is asking about Boston, MA so I will call the function with 'Boston, MA' <now call the function>"""
        ),
        UserMessage(
            content="What's the current weather like in Boston, MA today? First think step by step as to which city and state to call, only then call the get_current_weather function. "
        ),
    ]

    request = ChatCompletionRequest(
        model="gpt-4o-2024-11-20",
        messages=messages,
        tools=tools,
        max_completion_tokens=600,
        temperature=0.7,
    )
    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_structured_output(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant"),
        UserMessage(content="9.11 and 9.9 -- which is bigger?"),
    ]
    json_schema = {
        "name": "reasoning_schema",
        "strict": True,
        "schema": {
            "type": "object",
            "properties": {
                "reasoning_steps": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The reasoning steps leading to the final conclusion.",
                },
                "answer": {
                    "type": "string",
                    "description": "The final answer, taking into account the reasoning steps.",
                },
            },
            "required": ["reasoning_steps", "answer"],
            "additionalProperties": False,
        },
        "description": "A schema for structured output that includes reasoning steps and the final answer.",
    }

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=400,
        temperature=0.2,
        json_mode=False,
        structured_outputs=json_schema,
    )

    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_vision(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="Describe the animal in the image in one word."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        max_completion_tokens=200,
        temperature=0.5,
    )

    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_vision_tool_call(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(
            content="""You are detecting if there is text (numbers or letters) in images. 
If you see any text, call the ocr tool. It takes no parameters."""
        ),
        UserMessage(
            content=[
                ImageContent(
                    image_url=ImageUrl(
                        url=f"data:image/png;base64,{encode_image(numbers_image)}", detail=ImageDetail.LOW
                    )
                ),
            ]
        ),
    ]
    tools = [
        {
            "type": "function",
            "function": {
                "name": "ocr",
                "description": "Perform Optical Character Recognition (OCR) on an image",
                "parameters": {},
            },
        },
    ]

    request = ChatCompletionRequest(
        messages=messages,
        model="gpt-4o-2024-11-20",
        tools=tools,
        max_completion_tokens=200,
    )

    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_invalid_params(openai_aoai_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="gpt-4o-mini-2024-07-18",
        messages=[UserMessage(content="What is the capital of France?")],
        max_completion_tokens=100,
        context_window=1000,
        mirostat=1,
    )
    async for chunk in chat_completion_stream(request, "openai", openai_aoai_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


# endregion


# region Ollama
@pytest.fixture(
    params=[
        {"async_client": True},
    ]
)
def ollama_client_fixture(request: pytest.FixtureRequest) -> Callable[..., Any]:
    return ollama_client(**request.param)


async def test_chat_completion_stream_ollama(ollama_client_fixture: Callable[..., Any]) -> None:
    request = ChatCompletionRequest(
        model="llama3.2-vision:11b-instruct-q4_K_M",
        messages=[
            SystemMessage(content="Hello, world!"),
            UserMessage(content="What is the capital of France?"),
        ],
        max_completion_tokens=100,
        frequency_penalty=1.2,
        top_p=0.8,
        context_window=1000,
    )
    async for chunk in chat_completion_stream(request, "ollama", ollama_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_ollama_structured_output(ollama_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant"),
        UserMessage(content="9.11 and 9.9 -- which is bigger?"),
    ]
    json_schema = {
        "name": "reasoning_schema",
        "strict": True,
        "schema": {
            "type": "object",
            "properties": {
                "reasoning_steps": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The reasoning steps leading to the final conclusion.",
                },
                "answer": {
                    "type": "string",
                    "description": "The final answer, taking into account the reasoning steps.",
                },
            },
            "required": ["reasoning_steps", "answer"],
            "additionalProperties": False,
        },
        "description": "A schema for structured output that includes reasoning steps and the final answer.",
    }

    request = ChatCompletionRequest(
        messages=messages,
        model="llama3.2-vision:11b-instruct-q4_K_M",
        max_completion_tokens=400,
        temperature=0.2,
        json_mode=False,
        structured_outputs=json_schema,
    )

    async for chunk in chat_completion_stream(request, "ollama", ollama_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_ollama_multiple_tools(ollama_client_fixture: Callable[..., Any]) -> None:
    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_current_weather",
                "description": "Get the current weather",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        },
                        "format": {
                            "type": "string",
                            "enum": ["celsius", "fahrenheit"],
                            "description": "The temperature unit to use. Infer this from the users location.",
                        },
                    },
                    "required": ["location", "format"],
                },
            },
        },
    ]
    messages: list[MessageT] = [
        SystemMessage(content="Call the get_current_weather function once for each city that the user mentions."),
        UserMessage(content="What's the current weather like in Boston, MA and New York, NY today?"),
    ]
    request = ChatCompletionRequest(
        model="qwen2.5:14b",
        messages=messages,
        tools=tools,
        max_completion_tokens=400,
        temperature=0,
    )
    async for chunk in chat_completion_stream(request, "ollama", ollama_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_ollama_tool_message(ollama_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant"),
        UserMessage(content="What is the weather in Boston, MA?"),
        AssistantMessage(
            content="",
            tool_calls=[
                ToolCall(
                    id="abc123",
                    function=Function(
                        name="get_current_weather",
                        arguments={"location": "Boston, MA"},
                    ),
                )
            ],
        ),
        ToolMessage(name="abc123", content="The weather in Boston, MA is 70 degrees Fahrenheit."),
    ]
    request = ChatCompletionRequest(
        model="qwen2.5:14b",
        messages=messages,
        max_completion_tokens=300,
        temperature=0.3,
    )
    async for chunk in chat_completion_stream(request, "ollama", ollama_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


async def test_chat_completion_stream_ollama_vision_multiple_images(ollama_client_fixture: Callable[..., Any]) -> None:
    messages: list[MessageT] = [
        SystemMessage(content="You are a helpful assistant."),
        UserMessage(
            content=[
                TextContent(text="What are the animals in the images? Reply in one word for each animal."),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(cat_image)}", detail=ImageDetail.LOW)
                ),
                ImageContent(
                    image_url=ImageUrl(url=f"data:image/jpeg;base64,{encode_image(dog_image)}", detail=ImageDetail.LOW)
                ),
            ]
        ),
    ]
    request = ChatCompletionRequest(
        messages=messages,
        model="llama3.2-vision:11b-instruct-q4_K_M",
        max_completion_tokens=100,
    )
    async for chunk in chat_completion_stream(request, "ollama", ollama_client_fixture):
        print(chunk.model_dump(mode="json", exclude_none=True))


# endregion



from collections.abc import Callable
import os
import time
from typing import Any

from anthropic import Anthropic
from anthropic.types import Message

from not_again_ai.llm.chat_completion.types import (
    AssistantMessage,
    ChatCompletionChoice,
    ChatCompletionRequest,
    ChatCompletionResponse,
    Function,
    ToolCall,
)

ANTHROPIC_PARAMETER_MAP = {
    "max_completion_tokens": "max_tokens",
}


def anthropic_chat_completion(request: ChatCompletionRequest, client: Callable[..., Any]) -> ChatCompletionResponse:
    """Anthropic chat completion function.

    TODO
    - Image messages
    - Thinking
    - Citations
    - Stop sequences
    - Documents
    """
    kwargs = request.model_dump(mode="json", exclude_none=True)

    # For each key in ANTHROPIC_PARAMETER_MAP
    # If it is not None, set the key in kwargs to the value of the corresponding value in ANTHROPIC_PARAMETER_MAP
    # If it is None, remove that key from kwargs
    for key, value in ANTHROPIC_PARAMETER_MAP.items():
        if value is not None and key in kwargs:
            kwargs[value] = kwargs.pop(key)
        elif value is None and key in kwargs:
            del kwargs[key]

    # Handle messages
    # Any system messages need to be removed from messages and concatenated into a single string (in order)
    # Any tool messages need to be converted to a special user message
    # Any assistant messages with tool calls need to be converted.
    system = ""
    new_messages = []
    for message in kwargs["messages"]:
        if message["role"] == "system":
            system += message["content"] + "\n"
        elif message["role"] == "tool":
            new_messages.append(
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "tool_result",
                            "tool_use_id": message["name"],
                            "content": message["content"],
                        }
                    ],
                }
            )
        elif message["role"] == "assistant":
            content = []
            if message.get("content", None):
                content.append(
                    {
                        "type": "text",
                        "content": message["content"],
                    }
                )
            for tool_call in message.get("tool_calls", []):
                content.append(
                    {
                        "type": "tool_use",
                        "id": tool_call["id"],
                        "name": tool_call["function"]["name"],
                        "input": tool_call["function"]["arguments"],
                    }
                )
            new_messages.append(
                {
                    "role": "assistant",
                    "content": content,
                }
            )
        else:
            new_messages.append(message)
    kwargs["messages"] = new_messages
    system = system.strip()
    if system:
        kwargs["system"] = system

    # Handle tool choice and parallel tool calls
    if kwargs.get("tool_choice") is not None:
        tool_choice_value = kwargs.pop("tool_choice")
        tool_choice = {}
        if tool_choice_value == "none":
            tool_choice["type"] = "none"
        elif tool_choice_value in ["auto", "any"]:
            tool_choice["type"] = "auto"
            if kwargs.get("parallel_tool_calls") is not None:
                tool_choice["disable_parallel_tool_use"] = str(not kwargs["parallel_tool_calls"])
        else:
            tool_choice["name"] = tool_choice_value
            tool_choice["type"] = "tool"
            if kwargs.get("parallel_tool_calls") is not None:
                tool_choice["disable_parallel_tool_use"] = str(not kwargs["parallel_tool_calls"])
        kwargs["tool_choice"] = tool_choice
    kwargs.pop("parallel_tool_calls", None)

    start_time = time.time()
    response: Message = client(**kwargs)
    end_time = time.time()
    response_duration = round(end_time - start_time, 4)

    tool_calls: list[ToolCall] = []
    assistant_message = ""
    for block in response.content:
        if block.type == "text":
            assistant_message += block.text
        elif block.type == "tool_use":
            tool_calls.append(
                ToolCall(
                    id=block.id,
                    function=Function(
                        name=block.name,
                        arguments=block.input,  # type: ignore
                    ),
                )
            )

    choice = ChatCompletionChoice(
        message=AssistantMessage(
            content=assistant_message,
            tool_calls=tool_calls,
        ),
        finish_reason=response.stop_reason or "stop",
    )

    chat_completion_response = ChatCompletionResponse(
        choices=[choice],
        errors="",
        completion_tokens=response.usage.output_tokens,
        prompt_tokens=response.usage.input_tokens,
        cache_read_input_tokens=response.usage.cache_read_input_tokens,
        cache_creation_input_tokens=response.usage.cache_creation_input_tokens,
        response_duration=response_duration,
    )
    return chat_completion_response


def create_client_callable(client_class: type[Anthropic], **client_args: Any) -> Callable[..., Any]:
    """Creates a callable that instantiates and uses an Anthropic client.

    Args:
        client_class: The Anthropic client class to instantiate
        **client_args: Arguments to pass to the client constructor

    Returns:
        A callable that creates a client and returns completion results
    """
    filtered_args = {k: v for k, v in client_args.items() if v is not None}

    def client_callable(**kwargs: Any) -> Any:
        client = client_class(**filtered_args)
        completion = client.beta.messages.create(**kwargs)
        return completion

    return client_callable


def anthropic_client(api_key: str | None = None) -> Callable[..., Any]:
    if not api_key:
        api_key = os.environ.get("ANTHROPIC_API_KEY")
    client_callable = create_client_callable(Anthropic, api_key=api_key)
    return client_callable



from collections.abc import Generator

from playwright.sync_api import Browser, Playwright
import pytest

from not_again_ai.data.web import create_browser, get_raw_web_content


@pytest.fixture
def browser_fixture() -> Generator[tuple[Playwright, Browser], None, None]:
    p, browser = create_browser(headless=True)
    yield p, browser
    browser.close()
    p.stop()


def test_get_raw_web_content_new_browser() -> None:
    content = get_raw_web_content("https://playwright.dev/")
    assert content, "Content should not be empty"


def test_get_raw_web_content_existing_browser(browser_fixture: tuple[Playwright, Browser]) -> None:
    _, browser = browser_fixture
    content = get_raw_web_content(
        "https://playwright.dev/",
        browser=browser,
    )
    assert content, "Content should not be empty"



from collections.abc import AsyncGenerator, Callable
from typing import Any

from not_again_ai.llm.chat_completion.providers.anthropic_api import anthropic_chat_completion
from not_again_ai.llm.chat_completion.providers.ollama_api import ollama_chat_completion, ollama_chat_completion_stream
from not_again_ai.llm.chat_completion.providers.openai_api import openai_chat_completion, openai_chat_completion_stream
from not_again_ai.llm.chat_completion.types import ChatCompletionChunk, ChatCompletionRequest, ChatCompletionResponse


def chat_completion(
    request: ChatCompletionRequest,
    provider: str,
    client: Callable[..., Any],
) -> ChatCompletionResponse:
    """Get a chat completion response from the given provider. Currently supported providers:
    - `openai` - OpenAI
    - `azure_openai` - Azure OpenAI
    - `ollama` - Ollama

    Args:
        request: Request parameter object
        provider: The supported provider name
        client: Client information, see the provider's implementation for what can be provided

    Returns:
        ChatCompletionResponse: The chat completion response.
    """
    if provider == "openai" or provider == "azure_openai":
        return openai_chat_completion(request, client)
    elif provider == "ollama":
        return ollama_chat_completion(request, client)
    elif provider == "anthropic":
        return anthropic_chat_completion(request, client)
    else:
        raise ValueError(f"Provider {provider} not supported")


async def chat_completion_stream(
    request: ChatCompletionRequest,
    provider: str,
    client: Callable[..., Any],
) -> AsyncGenerator[ChatCompletionChunk, None]:
    """Stream a chat completion response from the given provider. Currently supported providers:
    - `openai` - OpenAI
    - `azure_openai` - Azure OpenAI
    - `ollama` - Ollama
    - `anthropic` - Anthropic

    Args:
        request: Request parameter object
        provider: The supported provider name
        client: Client information, see the provider's implementation for what can be provided

    Returns:
        AsyncGenerator[ChatCompletionChunk, None]
    """
    request.stream = True
    if provider == "openai" or provider == "azure_openai":
        async for chunk in openai_chat_completion_stream(request, client):
            yield chunk
    elif provider == "ollama":
        async for chunk in ollama_chat_completion_stream(request, client):
            yield chunk
    else:
        raise ValueError(f"Provider {provider} not supported")



from collections.abc import Collection, Set
from typing import Literal

from loguru import logger
import tiktoken

from not_again_ai.llm.chat_completion.types import MessageT
from not_again_ai.llm.prompting.types import BaseTokenizer


class TokenizerOpenAI(BaseTokenizer):
    def __init__(
        self,
        model: str,
        provider: str = "openai",
        allowed_special: Literal["all"] | Set[str] | None = None,
        disallowed_special: Literal["all"] | Collection[str] | None = None,
    ):
        self.model = model
        self.provider = provider
        self.allowed_special = allowed_special
        self.disallowed_special = disallowed_special

        self.init_tokenizer(model, provider, allowed_special, disallowed_special)

    def init_tokenizer(
        self,
        model: str,
        provider: str = "openai",
        allowed_special: Literal["all"] | Set[str] | None = None,
        disallowed_special: Literal["all"] | Collection[str] | None = None,
    ) -> None:
        try:
            self.encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            default_encoding = "o200k_base"
            logger.warning(f"Model {model} not found. Using {default_encoding} encoding.")
            self.encoding = tiktoken.get_encoding(default_encoding)

        # Set defaults if not provided
        if not allowed_special:
            self.allowed_special = set()
        if not disallowed_special:
            self.disallowed_special = ()

    def truncate_str(self, text: str, max_len: int) -> str:
        tokens = self.encoding.encode(
            text,
            allowed_special=self.allowed_special if self.allowed_special is not None else set(),
            disallowed_special=self.disallowed_special if self.disallowed_special is not None else (),
        )
        if len(tokens) > max_len:
            tokens = tokens[:max_len]
            truncated_text = self.encoding.decode(tokens)
            return truncated_text
        else:
            return text

    def num_tokens_in_str(self, text: str) -> int:
        return len(
            self.encoding.encode(
                text,
                allowed_special=self.allowed_special if self.allowed_special is not None else set(),
                disallowed_special=self.disallowed_special if self.disallowed_special is not None else (),
            )
        )

    def num_tokens_in_messages(self, messages: list[MessageT]) -> int:
        if self.model in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-3.5-turbo-1106",
            "gpt-3.5-turbo-0125",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
            "gpt-4-1106-preview",
            "gpt-4-turbo-preview",
            "gpt-4-0125-preview",
            "gpt-4-turbo",
            "gpt-4-turbo-2024-04-09",
            "gpt-4o",
            "gpt-4o-2024-05-13",
            "gpt-4o-2024-08-06",
            "gpt-4o-2024-11-20",
            "gpt-4o-mini",
            "gpt-4o-mini-2024-07-18",
            "o1",
            "o1-2024-12-17",
            "o1-mini",
            "o1-mini-2024-09-12",
            "o1-preview",
            "o1-preview-2024-09-12",
        }:
            tokens_per_message = 3  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_name = 1  # if there's a name, the role is omitted
        elif self.model == "gpt-3.5-turbo-0301":
            tokens_per_message = 4
            tokens_per_name = -1
        else:
            logger.warning(f"Model {self.model} not supported. Assuming gpt-4o encoding.")
            tokens_per_message = 3
            tokens_per_name = 1

        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            message_dict = message.model_dump(exclude_none=True)
            for key, value in message_dict.items():
                if isinstance(value, str):
                    num_tokens += len(
                        self.encoding.encode(
                            value,
                            allowed_special=self.allowed_special if self.allowed_special is not None else set(),
                            disallowed_special=self.disallowed_special if self.disallowed_special is not None else (),
                        )
                    )
                    if key == "name":
                        num_tokens += tokens_per_name
        num_tokens += 3
        return num_tokens



import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import seaborn as sns

from not_again_ai.base.file_system import create_file_dir
from not_again_ai.viz.utils import reset_plot_libs


def univariate_distplot(
    data: list[float] | npt.NDArray[np.float64],
    save_pathname: str,
    print_summary: bool = True,
    title: str | None = None,
    xlabel: str | None = "Value",
    ylabel: str | None = "Count",
    xlim: tuple[float, float] | None = None,
    ylim: tuple[float, float] | None = None,
    xticks: npt.ArrayLike | None = None,
    yticks: npt.ArrayLike | None = None,
    bins: int = 50,
    font_size: float = 48,
    height: float = 13,
    aspect: float = 2.2,
) -> None:
    """Saves a univariate distribution plot to the specified pathname.

    Args:
        data (listlike): Input listlike data to plot distribution of
        save_pathname (str): Filepath to save plot to. Parent directories will be automatically created.
        print_summary (bool, optional): If true will print summary statistics. Defaults to True.
        title (str, optional): Title of the plot. Defaults to None.
        xlabel (str, optional): Set the label for the x-axis. Defaults to 'Value'.
        ylabel (str, optional): Set the label for the y-axis. Defaults to 'Count'.
        xlim (tuple[float, float], optional): Set the x-axis limits (lower, upper). Defaults to None.
        ylim (tuple[float, float], optional): Set the y-axis limits (lower, upper). Defaults to None.
        xticks (npt.ArrayLike, optional): Set the x-axis tick locations. Defaults to None.
        yticks (npt.ArrayLike, optional): Set the y-axis tick locations. Defaults to None.
        bins (int, optional): See matplotlib [histplot documentation](https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.hist.html#matplotlib-pyplot-hist) for all options. Defaults to 50.
        font_size (float, optional): Font size. Defaults to 48.
        height (float, optional): Height (in inches) of the plot. Defaults to 13.
        aspect (float, optional): Aspect ratio of the plot, so that `aspect` * `height` gives the width of each facet in inches. Defaults to 2.2.
    """

    sns.set_theme(
        style="white",
        rc={
            "font.size": font_size,
            "axes.titlesize": font_size,
            "axes.labelsize": font_size * 0.8,
            "xtick.labelsize": font_size * 0.7,
            "ytick.labelsize": font_size * 0.7,
            "legend.fontsize": font_size * 0.55,
        },
    )

    # precompute summary statistics
    mean = np.mean(data)
    median = np.median(data)
    stdev = np.std(data)
    percentile_5 = np.percentile(data, 5)
    percentile_95 = np.percentile(data, 95)

    facet_grid = sns.displot(data, bins=bins, height=height, aspect=aspect)

    facet_grid.set(
        xlim=xlim,
        ylim=ylim,
        title=title,
        xlabel=xlabel,
        ylabel=ylabel,
    )

    if xticks is not None:
        facet_grid.set(xticks=xticks)
    if yticks is not None:
        facet_grid.set(yticks=yticks)

    pastel_colors = sns.color_palette("pastel")

    ax = facet_grid.axes.flatten()[0]

    # plot summary statistic lines
    ax.axvline(x=mean, color=pastel_colors[1], ls="--", lw=2.5, label=f"Mean: {mean:.3f}")
    ax.axvline(x=median, color=pastel_colors[2], ls="--", lw=2.5, label=f"Median: {mean:.3f}")
    ax.axvline(x=percentile_5, color=pastel_colors[9], ls="--", lw=2.5, label=f"5 Percentile: {percentile_5:.3f}")
    ax.axvline(x=percentile_95, color=pastel_colors[4], ls="--", lw=2.5, label=f"95 Percentile: {percentile_95:.3f}")
    # add legend for these lines
    handles, _ = ax.get_legend_handles_labels()
    # and an empty patch for the stdev statistic
    handles.append(mpatches.Patch(color="none", label=f"St Dev: {stdev:.3f}"))

    plt.legend(handles=handles, loc=0)

    create_file_dir(save_pathname)
    plt.savefig(save_pathname, bbox_inches="tight")
    reset_plot_libs()

    if print_summary:
        to_print = (
            "Summary Statistics:",
            f"Mean:\t\t{mean:.3f}",
            f"Median:\t\t{median:.3f}",
            f"5 Percentile:\t{percentile_5:.3f}",
            f"95 Percentile:\t{percentile_95:.3f}",
            f"St Dev:\t\t{stdev:.3f}",
        )
        print("\n".join(to_print))
